#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 14 17:55:39 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import os
import numpy as np
import matplotlib.pyplot as plt
from transitleastsquares import transitleastsquares as tls
from transitleastsquares import transit_mask, cleaned_array, catalog_info

#::: my modules
try:
    from exoworlds.tess import tessio
except:
    pass



###############################################################################
#::: helper functiomns
###############################################################################

#::: apply a mask (if wished so)
def mask(time, flux, period, duration, T0):
    intransit = transit_mask(time, period, duration, T0)
    time = time[~intransit]
    flux = flux[~intransit]
    time, flux = cleaned_array(time, flux)
    return time, flux
    


###############################################################################
#::: TLS search on an input lightcurve
###############################################################################
def tls_search(time, flux, flux_err,                                           #lightcurve
               SNR_threshold=5.,                                               #mine
               known_transits=None,                                            #mine
               mask_multiplier=1.5,                                            #mine
               show_plot=False, save_plot=False, outdir='',                    #mine
               R_host=1., R_host_min=0.13, R_host_max=3.5,                     #tls
               M_host=1., M_host_min=0.1, M_host_max=1.,                       #tls
               ldc=[0.4804, 0.1867],                                           #tls
               n_transits_min=3,                                               #tls
               **tls_kwargs):                                                  #tls
    '''
    Summary:
    -------
    runs TLS on these data with the given infos
    
    Inputs:
    -------
    time : array of flaot
        time stamps of observations
    flux : array of flaot
        normalized flux
    flux_err : array of flaot
        error of normalized flux
        
    Optional Inputs:
    ----------------
    SNR_threshold : float
        the SNR threshold at which to stop the TLS search
    known_transits : None or dict
        if dict and one transit is already known: 
            known_transits = {'period':[1.3], 'duration':[2.1], 'epoch':[245800.0]}
        if dict and multiple transits are already known: 
            known_transits = {'period':[1.3, 21.0], 'duration':[2.1, 4.1], 'epoch':[245800.0, 245801.0]}
        'period' is the period of the transit
        'duration' must be the total duration, i.e. from first ingress point to last egrees point, in days
        'epoch' is the epoch of the transit
    R_host : float
        radius of the star (e.g. median)
        default 1 R_sun (from TLS)
    R_host_min : float
        minimum radius of the star (e.g. 1st percentile)
        default 0.13 R_sun (from TLS)
    R_host_max : float
        maximum radius of the star (e.g. 99th percentile)
        default 3.5 R_sun (from TLS)
    M_host : float
        mass of the star (e.g. median)
        default 1. M_sun (from TLS)
    M_host_min : float
        minimum mass of the star (e.g. 1st percentile)
        default 0.1 M_sun (from TLS)
    M_host_max : float
        maximum mass of the star (e.g. 99th percentile)
        default 1. M_sun (from TLS)    
    ldc : list
        quadratic limb darkening parameters
        default [0.4804, 0.1867]
    show_plot : bool
        show a plot of each phase-folded transit candidate and TLS model in the terminal 
    show_plot : bool
        save a plot of each phase-folded transit candidate and TLS model into outdir
    outdir : string
        if None, use the current working directory
    **tls_kwargs : keyword arguments
        any other TLS keyword arguments, e.g. period_min=0.4
        
    Returns:
    -------
    list of all TLS results
    '''
    
    SNR = 1e12
    FOUND_SIGNAL = False
    results_all = []    
    if len(outdir)>0 and not os.path.exists(outdir): os.makedirs(outdir)    
    
    #::: apply a mask (if wished so)
    if known_transits is not None:
        for period, duration, T0 in zip(known_transits['period'], known_transits['duration'], known_transits['epoch']):
            time, flux = mask(time, flux, period, duration, T0)
    
    #::: search for the rest
    i = 0
    while (SNR >= SNR_threshold) and (FOUND_SIGNAL==False):
        model = tls(time, flux, flux_err)
        results = model.power(R_star=R_host, R_star_min=R_host_min, R_star_max=R_host_max, 
                              M_star=M_host, M_star_min=M_host_min, M_star_max=M_host_max,
                              u=ldc, 
                              n_transits_min=n_transits_min, 
                              show_progress_bar=False,
                              **tls_kwargs)
        
        if results.snr >= SNR_threshold:
            time, flux = mask(time, flux, results.period, mask_multiplier*results.duration, results.T0)
            results_all.append(results)
        
            if show_plot or save_plot:
                fig, ax = plt.subplots()
                ax.plot(results['folded_phase'], results['folded_y'], 'b.')
                ax.plot(results['model_folded_phase'], results['model_folded_model'], 'r-', lw=2)
                ax.text( 1.02, 0.95, 'Period = ' + np.format_float_positional(results['period'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
                ax.text( 1.02, 0.85, 'Depth = ' + np.format_float_positional(1e3*(1.-results['depth']),4) + ' ppt', ha='left', va='center', transform=ax.transAxes )
                ax.text( 1.02, 0.75, 'Duration = ' + np.format_float_positional(24*(1.-results['duration']),4) + ' h', ha='left', va='center', transform=ax.transAxes )
                ax.text( 1.02, 0.65, 'Epoch = ' + np.format_float_positional(results['T0'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
                ax.text( 1.02, 0.55, 'SNR = ' + np.format_float_positional(results['snr'],4), ha='left', va='center', transform=ax.transAxes )
                if save_plot:
                    fig.savefig(os.path.join(outdir,'tls_signal_'+str(i)), bbox_inches='tight')
                if show_plot:
                    plt.show(fig)
                else:
                    plt.close(fig)
            
        SNR = results.snr
        i+=1
                
    return results_all



###############################################################################
#::: TLS search using tessio
###############################################################################
def tls_search_by_tic(tic_id, sigma_multiplier=3, **kwargs):
    '''
    Summary:
    -------
    - wrapper around tls_search()
    - retrieves the SPOC PDC-SAP lightcurve
    - retrieves all TIC catalog information from MAST
    - runs TLS on these data and infos
        
    Inputs:
    -------
    tic_id : str
        TIC ID
        
    Optional Inputs:
    ----------------
    sigma_multiplier : float 
        set TLS' uniform bounds for the host's radius and mass at sigma_multiplier times the error bars from TICv8
        default is 3
    **kwargs : keyword arguments
        any other keyword arguments, see tls_search_by_tic()
        
    
    Returns:
    -------
        - list of all TLS results
    '''
    
    #::: format inputs
    tic_id = str(int(tic_id))
    
    #::: load data and inject transit
    time, flux, flux_err = tessio.get(tic_id, pipeline='spoc', PDC=True, unpack=True)
    
    #::: load TIC info
    ldc, R_host, R_host_lerr, R_host_uerr, M_host, M_host_lerr, M_host_uerr = catalog_info(TIC_ID=int(tic_id))
    print('\nTICv8 info:')
    print('Quadratic limb darkening u_0, u_1', ldc[0], ldc[1])
    print('Stellar radius', R_host, '+', R_host_lerr, '-', R_host_uerr)
    print('Stellar mass', M_host, '+', M_host_lerr, '-', M_host_uerr)
    
    return tls_search(time, flux, flux_err,
                      R_host=R_host, R_host_min=R_host-sigma_multiplier*R_host_lerr, R_host_max=R_host+sigma_multiplier*R_host_uerr, 
                      M_host=M_host, M_host_min=M_host-sigma_multiplier*M_host_lerr, M_host_max=M_host+sigma_multiplier*M_host_uerr,
                      ldc=ldc,
                      **kwargs)



###############################################################################
#::: main
###############################################################################
if __name__ == '__main__':
    pass
    
    ###########################################################################
    #::: Example: search for a transit with TLS and tessio
    ###########################################################################
    # tic_id = '269701147'
    # SNR_threshold=5.,
    # known_transits = {'epoch':[2458715.3547, 2458726.0526, 2458743.5534],
    #                   'period':[8.8806, 28.5810, 38.3497],
    #                   'duration':[3.09/24., 4.45/24., 5.52/24.]
    #                  }
    
    # results_all = tls_search_by_tic(tic_id,
    #                 SNR_threshold=SNR_threshold,
    #                 known_transits=known_transits)
    # print(results_all)


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 28 15:19:30 2018

@author:
Maximilian N. Günther
MIT Kavli institute for Astrophysics and Space Research, 
Massachusetts institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import os
#import collections
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter
import pickle
from corner import corner
from tqdm import tqdm 
from astropy.constants import M_earth, M_jup, M_sun, R_earth, R_jup, R_sun, au
import copy
from multiprocessing import Pool
from contextlib import closing

#::: allesfitter modules
from . import config
from .utils.latex_printer import round_tex
from .general_output import logprint
from .priors.simulate_PDF import simulate_PDF
from .limb_darkening import LDC3
from .computer import update_params, calculate_model, flux_fct, flux_subfct_ellc, flux_subfct_sinusoidal_phase_curves
from .exoworlds_rdx.lightcurves.index_transits import index_transits
from .lightcurves.lightcurves import get_epoch_occ


###############################################################################
#::: constants (replaced with astropy.constants)
###############################################################################
#M_earth = 5.9742e+24 	#kg 	Earth mass
#M_jup   = 1.8987e+27 	#kg 	Jupiter mass
#M_sun   = 1.9891e+30 	#kg 	Solar mass
#R_earth = 6378136      #m 	Earth equatorial radius
#R_jup   = 71492000 	#m 	Jupiter equatorial radius
#R_sun   = 695508000 	#m 	Solar radius



###############################################################################
#::: globals
#::: sorry for that... it's multiprocessing, not me, I swear!
###############################################################################
# companion = None
# inst = None
# samples2 = None
# derived_samples = None


###############################################################################
#::: calculate values from model curves
###############################################################################
def calculate_values_from_model_curves(p, inst, companion):
    '''
    Parameters
    ----------
    p : dict
        parameters corresponding to one single sample
    inst : str
        instrument name
    companion : str
        companion name

    Returns
    -------
    list
        list containing the transit depth, occultation depth, and nightside flux
    '''
    
    #==========================================================================
    #::: init
    #==========================================================================
    depth_tr = np.nan
    depth_occ = np.nan
    nightside_flux = np.nan
    epoch_occ = get_epoch_occ(p[companion+'_epoch'], p[companion+'_period'], p[companion+'_f_s'], p[companion+'_f_c'])


    #==========================================================================
    #:: calculating
    #==========================================================================
    
    #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    #::: without phase curve
    #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    if (config.BASEMENT.settings['phase_curve'] is False):
        #::: compute transit / primary eclipse depth
        depth_tr = 1. - flux_subfct_ellc(p, inst, companion, xx=[p[companion+'_epoch']])[0]
        
        #::: compute occultation / secondary eclipse depth (if wished)
        if (config.BASEMENT.settings['secondary_eclipse'] is True): 
            depth_occ = 1. - flux_subfct_ellc(p, inst, companion, xx=[epoch_occ])[0]


    #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    #::: with phase curve sine_series or sine_physical
    #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    elif (config.BASEMENT.settings['phase_curve'] is True) and (config.BASEMENT.settings['phase_curve_style'] in ['sine_series','sine_physical']):
        
        #::: 0: epoch, 1: epoch_occ
        xx0 = [p[companion+'_epoch'], epoch_occ]
        
        #::: for debugging
        # xx0 = np.linspace(p[companion+'_epoch']-0.25*p[companion+'_period'], p[companion+'_epoch']+0.75*p[companion+'_period'], 1001)
        
        #::: the full model flux with phase curve and dips
        phase_curve_dips = flux_fct(p, inst, companion, xx=xx0)

        #::: the phase curve without any dips
        ellc_flux, ellc_flux1, ellc_flux2 = flux_subfct_ellc(p, inst, companion, xx=xx0, return_fluxes=True)
        phase_curve_no_dips = flux_subfct_sinusoidal_phase_curves(p, inst, companion, np.ones_like(xx0), xx=xx0)
        
        #::: the phase curve with atmopsheric dips, but without nightside flux (sbratio=1e-12)
        p2 = copy.deepcopy(p)
        p2[companion+'_sbratio_'+inst] = 1e-12
        ellc_flux, ellc_flux1, ellc_flux2 = flux_subfct_ellc(p2, inst, companion, xx=xx0, return_fluxes=True)
        phase_curve_atmo_dips = flux_subfct_sinusoidal_phase_curves(p2, inst, companion, ellc_flux2, xx=xx0)

        #::: for debugging
        # fig = plt.figure()
        # plt.plot(xx0, phase_curve_dips, label='phase_curve_dips')
        # plt.plot(xx0, phase_curve_no_dips, label='phase_curve_no_dips')
        # plt.plot(xx0, phase_curve_atmo_dips, label='phase_curve_atmo_dips')
        # plt.legend()
        # plt.ylim([0.999,1.001])
        # plt.axhline(1,c='grey',ls='--')
        # fig.savefig(os.path.join(config.BASEMENT.outdir,'phase_curve_depths.pdf'), bbox_inches='tight')

        #::: compute transit / primary eclipse depth
        depth_tr = 1e3 * (phase_curve_no_dips[0] - phase_curve_dips[0]) #in ppt; 0: epoch
        
        #::: compute
        if (config.BASEMENT.settings['secondary_eclipse'] is True): 
            
            #::: compute occultation / secondary eclipse depth
            depth_occ = 1e3 * (phase_curve_no_dips[1] - phase_curve_dips[1]) #in ppt; 1: epoch_occ
                
            #::: compute nightside flux
            nightside_flux = 1e3 * (phase_curve_atmo_dips[1] - phase_curve_dips[1]) #in ppt; 1: epoch_occ
            

    #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    #::: with phase curve ellc_physical
    #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    elif (config.BASEMENT.settings['phase_curve'] is True) and (config.BASEMENT.settings['phase_curve_style'] in ['ellc_physical']):
        pass #TODO: not yet implemented
        
    
    #==========================================================================
    #::: return
    #==========================================================================
    return [depth_tr, depth_occ, nightside_flux]




###############################################################################
#::: the main derive function
###############################################################################
def derive(samples, mode):
    '''
    Derives parameter of the system using Winn 2010
    
    Input:
    ------
    samples : array
        samples from the mcmc or nested sampling
    mode : str
        'mcmc' or 'ns'
        
    Returns:
    --------
    derived_samples : dict 
        with keys 'i', 'R1a', 'R2a', 'k', 'depth_undiluted', 'b_tra', 'b_occ', 'Ttot', 'Tfull'
        each key contains all the samples derived from the MCMC samples 
        (not mean values, but pure samples!)
        i = inclination 
        R1a = R1/a, radius companion over semiamplitude
        R2a = R2/a, radius star over semiamplitude
        Ttot = T_{1-4}, total transit width 
        Tfull = T_{2-3}, full transit width
        
    Output:
    -------
    latex table of results
    corner plot of derived values posteriors
    '''
    
    #::: using a global keyword 
    #::: sorry for that... it's multiprocessing, not me, I swear!
    global companion
    global inst
    global samples2
    global derived_samples
    
    samples2 = samples #global variable
    N_samples = samples.shape[0]
    

    #==========================================================================
    #::: stellar 'posteriors'
    #==========================================================================
    if os.path.exists( os.path.join(config.BASEMENT.datadir,'params_star.csv') ):
        buf = np.genfromtxt( os.path.join(config.BASEMENT.datadir,'params_star.csv'), delimiter=',', names=True, dtype=None, encoding='utf-8', comments='#' )
        star = {}
        star['R_star'] = simulate_PDF(buf['R_star'], buf['R_star_lerr'], buf['R_star_uerr'], size=N_samples, plot=False)
        star['M_star'] = simulate_PDF(buf['M_star'], buf['M_star_lerr'], buf['M_star_uerr'], size=N_samples, plot=False)
        star['Teff_star'] = simulate_PDF(buf['Teff_star'], buf['Teff_star_lerr'], buf['Teff_star_uerr'], size=N_samples, plot=False)
    else:
        star = {'R_star':np.nan, 'M_star':np.nan, 'Teff_star':np.nan}
    
    
    #==========================================================================
    #::: derive all the params
    #==========================================================================
    companions = config.BASEMENT.settings['companions_all']
    
    def get_params(key):
        ind = np.where(config.BASEMENT.fitkeys==key)[0]
        if len(ind)==1: 
            return samples[:,ind].flatten() #if it was fitted for
        else: 
            try:
                if config.BASEMENT.params[key] is None:
                    return np.nan #if None, retun nan instead
                else:
                    return config.BASEMENT.params[key] #else take the input value
            except KeyError:
                return np.nan #if all fails, return nan
        
    def sin_d(alpha): return np.sin(np.deg2rad(alpha))
    def cos_d(alpha): return np.cos(np.deg2rad(alpha))
    def arcsin_d(x): return np.rad2deg(np.arcsin(x))
    def arccos_d(x): return np.rad2deg(np.arccos(x))

    derived_samples = {}
    for cc in companions:
        companion = cc
        
        #----------------------------------------------------------------------
        #::: radii
        #----------------------------------------------------------------------
        derived_samples[companion+'_R_star/a'] = get_params(companion+'_rsuma') / (1. + get_params(companion+'_rr'))
        derived_samples[companion+'_a/R_star'] = (1. + get_params(companion+'_rr')) / get_params(companion+'_rsuma')
        derived_samples[companion+'_R_companion/a'] = get_params(companion+'_rsuma') * get_params(companion+'_rr') / (1. + get_params(companion+'_rr'))
        derived_samples[companion+'_R_companion_(R_earth)'] = star['R_star'] * get_params(companion+'_rr') * R_sun.value / R_earth.value #in R_earth
        derived_samples[companion+'_R_companion_(R_jup)'] = star['R_star'] * get_params(companion+'_rr') * R_sun.value / R_jup.value #in R_jup

    
        #----------------------------------------------------------------------
        #::: orbit
        #----------------------------------------------------------------------
        derived_samples[companion+'_a_(R_sun)'] = star['R_star'] / derived_samples[companion+'_R_star/a']   
        derived_samples[companion+'_a_(AU)'] = derived_samples[companion+'_a_(R_sun)'] * R_sun.value/au.value
        derived_samples[companion+'_i'] = arccos_d(get_params(companion+'_cosi')) #in deg
        derived_samples[companion+'_e'] = get_params(companion+'_f_s')**2 + get_params(companion+'_f_c')**2
        derived_samples[companion+'_e_sinw'] = get_params(companion+'_f_s') * np.sqrt(derived_samples[companion+'_e'])
        derived_samples[companion+'_e_cosw'] = get_params(companion+'_f_c') * np.sqrt(derived_samples[companion+'_e'])
        derived_samples[companion+'_w'] = np.rad2deg(np.mod( np.arctan2(get_params(companion+'_f_s'), get_params(companion+'_f_c')), 2*np.pi) ) #in deg, from 0 to 360
        if np.isnan(derived_samples[companion+'_w']).all():
            derived_samples[companion+'_w'] = 0.
        
        
        #----------------------------------------------------------------------
        #::: masses
        #----------------------------------------------------------------------
        if companion+'_K' in config.BASEMENT.params:
            a_1 = 0.019771142 * get_params(companion+'_K') * get_params(companion+'_period') * np.sqrt(1. - derived_samples[companion+'_e']**2)/sin_d(derived_samples[companion+'_i'])
    #        derived_samples[companion+'_a_rv'] = (1.+1./ellc_params[companion+'_q'])*a_1
            derived_samples[companion+'_q'] = 1./(( derived_samples[companion+'_a_(R_sun)'] / a_1 ) - 1.)
            derived_samples[companion+'_M_companion_(M_earth)'] = derived_samples[companion+'_q'] * star['M_star'] * M_sun.value / M_earth.value #in M_earth
            derived_samples[companion+'_M_companion_(M_jup)'] = derived_samples[companion+'_q'] * star['M_star'] * M_sun.value / M_jup.value #in M_jup
            
            
        #----------------------------------------------------------------------
        #::: time of secondary eclipse   
        #---------------------------------------------------------------------- 
        if config.BASEMENT.settings['secondary_eclipse'] is True:
            derived_samples[companion+'_epoch_occ'] = get_params(companion+'_epoch') + get_params(companion+'_period')/2. * (1. + 4./np.pi * derived_samples[companion+'_e'] * cos_d(derived_samples[companion+'_w'])  ) #approximation from Winn2010
        
        
        #----------------------------------------------------------------------
        #::: impact params of primary eclipse with eccentricity corrections (from Winn 2010) 
        #----------------------------------------------------------------------
        eccentricity_correction_b_tra = ( (1. - derived_samples[companion+'_e']**2) / ( 1. + derived_samples[companion+'_e']*sin_d(derived_samples[companion+'_w']) ) )
        
        derived_samples[companion+'_b_tra'] = (1./derived_samples[companion+'_R_star/a']) * get_params(companion+'_cosi') * eccentricity_correction_b_tra
        
        
        #----------------------------------------------------------------------
        #::: impact params of secondary eclipse with eccentricity corrections (from Winn 2010) 
        #----------------------------------------------------------------------        
        eccentricity_correction_b_occ = ( (1. - derived_samples[companion+'_e']**2) / ( 1. - derived_samples[companion+'_e']*sin_d(derived_samples[companion+'_w']) ) )
        
        if config.BASEMENT.settings['secondary_eclipse'] is True:
            derived_samples[companion+'_b_occ'] = (1./derived_samples[companion+'_R_star/a']) * get_params(companion+'_cosi') * eccentricity_correction_b_occ
        
        
        #----------------------------------------------------------------------
        #::: transit duration (in hours) with eccentricity corrections (from Winn 2010) 
        #----------------------------------------------------------------------
        eccentricity_correction_T_tra = ( np.sqrt(1. - derived_samples[companion+'_e']**2) / ( 1. + derived_samples[companion+'_e']*sin_d(derived_samples[companion+'_w']) ) )
        
        derived_samples[companion+'_T_tra_tot'] = get_params(companion+'_period')/np.pi *24.  \
                                                  * np.arcsin( derived_samples[companion+'_R_star/a'] \
                                                               * np.sqrt( (1. + get_params(companion+'_rr'))**2 - derived_samples[companion+'_b_tra']**2 ) \
                                                               / sin_d(derived_samples[companion+'_i']) ) \
                                                  * eccentricity_correction_T_tra    #in h
        derived_samples[companion+'_T_tra_full'] = get_params(companion+'_period')/np.pi *24.  \
                                                   * np.arcsin( derived_samples[companion+'_R_star/a'] \
                                                                * np.sqrt( (1. - get_params(companion+'_rr'))**2 - derived_samples[companion+'_b_tra']**2  )\
                                                                / sin_d(derived_samples[companion+'_i']) ) \
                                                   * eccentricity_correction_T_tra    #in h
                                  

        #----------------------------------------------------------------------
        #::: primary and secondary eclipse depths (per inst) 
        #::: / transit and occultation depths (per inst)
        #----------------------------------------------------------------------
        for ii in config.BASEMENT.settings['inst_phot']:
            
            
            #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
            #::: setup
            #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
            inst = ii
            N_less_samples = 1000
            derived_samples[companion+'_depth_tr_dil_'+inst] = np.nan*np.empty(N_less_samples)
            derived_samples[companion+'_depth_occ_dil_'+inst] = np.nan*np.empty(N_less_samples)
            derived_samples[companion+'_nightside_flux_dil_'+inst] = np.nan*np.empty(N_less_samples)

            
            #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
            #::: iterate through all samples, draw different models and measure the depths
            #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
            logprint('\n' + inst + ' ' + companion + ': deriving eclipse depths and more from model curves...')
            for i in range(N_less_samples):
                s = samples[ np.random.randint(low=0,high=samples2.shape[0]) , : ]
                p = update_params(s)
                r = calculate_values_from_model_curves(p, inst, companion)
                derived_samples[companion+'_depth_tr_dil_'+inst][i] = r[0]
                derived_samples[companion+'_depth_occ_dil_'+inst][i] = r[1]
                derived_samples[companion+'_nightside_flux_dil_'+inst][i] = r[2]
                
            
            #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
            #resize the arrays to match the true N_samples (by redrawing the 1000 values)
            #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
            derived_samples[companion+'_depth_tr_dil_'+inst] = np.resize(derived_samples[companion+'_depth_tr_dil_'+inst], N_samples)
            derived_samples[companion+'_depth_occ_dil_'+inst] = np.resize(derived_samples[companion+'_depth_occ_dil_'+inst], N_samples)
            derived_samples[companion+'_nightside_flux_dil_'+inst] = np.resize(derived_samples[companion+'_nightside_flux_dil_'+inst], N_samples)

    
        #----------------------------------------------------------------------
        #::: undiluted (per companion; per inst)
        #----------------------------------------------------------------------
        for inst in config.BASEMENT.settings['inst_phot']:
            dil = get_params('dil_'+inst)
            if all(np.atleast_1d(np.isnan(dil))): dil = 0
            derived_samples[companion+'_depth_tr_undil_'+inst] = derived_samples[companion+'_depth_tr_dil_'+inst] / (1. - dil) #in ppt
            derived_samples[companion+'_depth_occ_undil_'+inst] = derived_samples[companion+'_depth_occ_dil_'+inst] / (1. - dil) #in ppt
            derived_samples[companion+'_nightside_flux_undil_'+inst] = derived_samples[companion+'_nightside_flux_dil_'+inst] / (1. - dil) #in ppt

        
        #----------------------------------------------------------------------
        #::: equilibirum temperature
        #::: currently assumes Albedo of 0.3 and Emissivity of 1
        #----------------------------------------------------------------------
        albedo = 0.3
        emissivity = 1.
        derived_samples[companion+'_Teq'] = star['Teff_star']  * ( (1.-albedo)/emissivity )**0.25 * np.sqrt(derived_samples[companion+'_R_star/a'] / 2.)
        
        
        #----------------------------------------------------------------------
        #::: stellar density from orbit
        #----------------------------------------------------------------------
        if companion in config.BASEMENT.settings['companions_phot']:
            if all(np.atleast_1d(get_params(companion+'_rr'))<0.215443469): #see computer.py; get_params could return np.nan (float) or array; all(np.atleast_1d(...)) takes care of that
                derived_samples[companion+'_host_density'] = 3. * np.pi * (1./derived_samples[companion+'_R_star/a'])**3. / (get_params(companion+'_period')*86400.)**2 / 6.67408e-8 #in cgs
  
    
        #----------------------------------------------------------------------
        #::: companion densities
        #----------------------------------------------------------------------
        derived_samples[companion+'_density'] = ( (derived_samples[companion+'_M_companion_(M_earth)'] * M_earth) / (4./3. * np.pi * (derived_samples[companion+'_R_companion_(R_earth)'] * R_earth)**3 ) ).cgs.value #in cgs
        
        
        #----------------------------------------------------------------------
        #::: the companion's surface gravity (individual posterior distribution for each companion; via Southworth+ 2007)
        #----------------------------------------------------------------------
        try:
            derived_samples[companion+'_surface_gravity'] = 2. * np.pi / (get_params(companion+'_period')*86400.) * np.sqrt((1.-derived_samples[companion+'_e']**2)) * (get_params(companion+'_K')*1e5) / (derived_samples[companion+'_R_companion/a'])**2 / sin_d(derived_samples[companion+'_i'])
        except:
            pass
        
        
        #----------------------------------------------------------------------
        #::: period ratios (for ressonance studies)
        #----------------------------------------------------------------------
        if len(companions)>1:
            for other_companion in companions:
                if other_companion is not companion:
                    derived_samples[companion+'_period/'+other_companion+'_period'] = get_params(companion+'_period') / get_params(other_companion+'_period')
                        
                    
        #----------------------------------------------------------------------
        #::: limb darkening
        #----------------------------------------------------------------------
        for inst in config.BASEMENT.settings['inst_all']:
            
            #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
            #::: host
            #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
            if config.BASEMENT.settings['host_ld_law_'+inst] is None:
                pass
                
            elif config.BASEMENT.settings['host_ld_law_'+inst] == 'lin':
                derived_samples['host_ldc_u1_'+inst] = get_params('host_ldc_q1_'+inst)
                
            elif config.BASEMENT.settings['host_ld_law_'+inst] == 'quad':
                derived_samples['host_ldc_u1_'+inst] = 2 * np.sqrt(get_params('host_ldc_q1_'+inst)) * get_params('host_ldc_q2_'+inst)
                derived_samples['host_ldc_u2_'+inst] = np.sqrt(get_params('host_ldc_q1_'+inst)) * (1. - 2. * get_params('host_ldc_q2_'+inst))
                
            elif config.BASEMENT.settings['host_ld_law_'+inst] == 'sing':
                derived_samples['host_ldc_u1_'+inst] = np.nan*np.empty(N_samples)
                derived_samples['host_ldc_u2_'+inst] = np.nan*np.empty(N_samples)
                derived_samples['host_ldc_u3_'+inst] = np.nan*np.empty(N_samples)
                for i in range(N_samples):
                    u1, u2, u3 = LDC3.forward([get_params('host_ldc_q1_'+inst)[i], get_params('host_ldc_q2_'+inst)[i], get_params('host_ldc_q3_'+inst)[i]])
                    derived_samples['host_ldc_u1_'+inst][i] = u1
                    derived_samples['host_ldc_u2_'+inst][i] = u2
                    derived_samples['host_ldc_u3_'+inst][i] = u3
                
            else:
                raise ValueError("Currently only 'none', 'lin', 'quad' and 'sing' limb darkening are supported.")
            
        
    #==========================================================================
    #::: median stellar density
    #==========================================================================
    derived_samples['combined_host_density'] = []
    for companion in config.BASEMENT.settings['companions_phot']:
        try: derived_samples['combined_host_density'] = np.append(derived_samples['combined_host_density'], derived_samples[companion+'_host_density'])
        except: pass
    

    
    ###############################################################################
    #::: write keys for output
    ###############################################################################
    names = []
    labels = []
    for companion in companions:
            
        names.append( companion+'_R_star/a' )
        labels.append( 'Host radius over semi-major axis '+companion+'; $R_\star/a_\mathrm{'+companion+'}$' )
        
        names.append( companion+'_a/R_star' )
        labels.append( 'Semi-major axis '+companion+' over host radius; $a_\mathrm{'+companion+'}/R_\star$' )
        
        names.append( companion+'_R_companion/a'  )
        labels.append( 'Companion radius '+companion+' over semi-major axis '+companion+'; $R_\mathrm{'+companion+'}/a_\mathrm{'+companion+'}$' )
        
        names.append( companion+'_R_companion_(R_earth)' )
        labels.append( 'Companion radius '+companion+'; $R_\mathrm{'+companion+'}$ ($\mathrm{R_{\oplus}}$)' )
        
        names.append( companion+'_R_companion_(R_jup)' )
        labels.append( 'Companion radius '+companion+'; $R_\mathrm{'+companion+'}$ ($\mathrm{R_{jup}}$)' )
        
        names.append( companion+'_a_(R_sun)' )
        labels.append( 'Semi-major axis '+companion+'; $a_\mathrm{'+companion+'}$ ($\mathrm{R_{\odot}}$)' )
        
        names.append( companion+'_a_(AU)' )
        labels.append( 'Semi-major axis '+companion+'; $a_\mathrm{'+companion+'}$ (AU)' )
        
        names.append( companion+'_i' )
        labels.append( 'Inclination '+companion+'; $i_\mathrm{'+companion+'}$ (deg)' )
        
        names.append( companion+'_e' )
        labels.append( 'Eccentricity '+companion+'; $e_\mathrm{'+companion+'}$' )
        
        names.append( companion+'_w' )
        labels.append( 'Argument of periastron '+companion+'; $w_\mathrm{'+companion+'}$ (deg)' )
        
        names.append( companion+'_q' )
        labels.append( 'Mass ratio '+companion+'; $q_\mathrm{'+companion+'}$' )
        
        names.append( companion+'_M_companion_(M_earth)' )
        labels.append( 'Companion mass '+companion+'; $M_\mathrm{'+companion+'}$ ($\mathrm{M_{\oplus}}$)' )
        
        names.append( companion+'_M_companion_(M_jup)' )
        labels.append( 'Companion mass '+companion+'; $M_\mathrm{'+companion+'}$ ($\mathrm{M_{jup}}$)' )
        
        names.append( companion+'_b_tra' )
        labels.append( 'Impact parameter '+companion+'; $b_\mathrm{tra;'+companion+'}$' )
        
        names.append( companion+'_T_tra_tot'  )
        labels.append( 'Total transit duration '+companion+'; $T_\mathrm{tot;'+companion+'}$ (h)' )
        
        names.append( companion+'_T_tra_full' )
        labels.append( 'Full-transit duration '+companion+'; $T_\mathrm{full;'+companion+'}$ (h)' )
        
        names.append( companion+'_epoch_occ'  )
        labels.append( 'Epoch occultation '+companion+'; $T_\mathrm{0;occ;'+companion+'}$' )
        
        names.append( companion+'_b_occ'  )
        labels.append( 'Impact parameter occultation '+companion+'; $b_\mathrm{occ;'+companion+'}$' )
        
        names.append( companion+'_host_density' )
        labels.append( 'Host density from orbit '+companion+'; $\\rho_\mathrm{\star;'+companion+'}$ (cgs)' )
    
        names.append( companion+'_density' )
        labels.append( 'Companion density '+companion+'; \\$rho_\mathrm{'+companion+'}$ (cgs)' )
        
        names.append( companion+'_surface_gravity')
        labels.append( 'Companion surface gravity '+companion+'; $g_\mathrm{'+companion+'}$ (cgs)' )
        
        names.append( companion+'_Teq' )
        labels.append( 'Equilibrium temperature '+companion+'; $T_\mathrm{eq;'+companion+'}$ (K)' )
        
        for inst in config.BASEMENT.settings['inst_phot']:
            
            names.append( companion+'_depth_tr_undil_'+inst )
            labels.append( 'Transit depth (undil.) '+companion+'; $\delta_\mathrm{tr; undil; '+companion+'; '+inst+'}$ (ppt)' )
            
            names.append( companion+'_depth_tr_dil_'+inst )
            labels.append( 'Transit depth (dil.) '+companion+'; $\delta_\mathrm{tr; dil; '+companion+'; '+inst+'}$ (ppt)' )
        
            names.append( companion+'_depth_occ_undil_'+inst )
            labels.append( 'Occultation depth (undil.) '+companion+'; $\delta_\mathrm{occ; undil; '+companion+'; '+inst+'}$ (ppt)' )
            
            names.append( companion+'_depth_occ_dil_'+inst )
            labels.append( 'Occultation depth (dil.) '+companion+'; $\delta_\mathrm{occ; dil; '+companion+'; '+inst+'}$ (ppt)' )
            
            names.append( companion+'_nightside_flux_undil_'+inst )
            labels.append( 'Nightside flux (undil.)'+companion+'; $F_\mathrm{nightside; undil; '+companion+'; '+inst+'}$ (ppt)' )
            
            names.append( companion+'_nightside_flux_dil_'+inst )
            labels.append( 'Nightside flux (dil.)'+companion+'; $F_\mathrm{nightside; dil; '+companion+'; '+inst+'}$ (ppt)' )
            
            
            
        #::: period ratios (for ressonance studies)
        if len(companions)>1:
            for other_companion in companions:
                if other_companion is not companion:
                    names.append( companion+'_period/'+other_companion+'_period' )
                    labels.append( 'Period ratio; $P_\mathrm{'+companion+'} / P_\mathrm{'+other_companion+'}$' )
           
            
    #::: host
    for inst in config.BASEMENT.settings['inst_all']:    
        if config.BASEMENT.settings['host_ld_law_'+inst] is None:
            pass
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'lin':
            names.append( 'host_ldc_u1_'+inst )
            labels.append( 'Limb darkening; $u_\mathrm{1; '+inst+'}$' )
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'quad':
            names.append( 'host_ldc_u1_'+inst )
            labels.append( 'Limb darkening; $u_\mathrm{1; '+inst+'}$' )
            names.append( 'host_ldc_u2_'+inst )
            labels.append( 'Limb darkening; $u_\mathrm{2; '+inst+'}$' )
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'sing':
            names.append( 'host_ldc_u1_'+inst )
            labels.append( 'Limb darkening; $u_\mathrm{1; '+inst+'}$' )
            names.append( 'host_ldc_u2_'+inst )
            labels.append( 'Limb darkening; $u_\mathrm{2; '+inst+'}$' )
            names.append( 'host_ldc_u3_'+inst )
            labels.append( 'Limb darkening; $u_\mathrm{3; '+inst+'}$' )
            
        else:
            raise ValueError("Currently only 'none', 'lin', 'quad' and 'sing' limb darkening are supported.")
                
        
    names.append( 'combined_host_density' )
    labels.append( 'Combined host density from all orbits; $rho_\mathrm{\star; combined}$ (cgs)' )
        
            
    ###############################################################################
    #::: delete pointless values
    ###############################################################################
    ind_good = []
    for i,name in enumerate(names):
        if (name in derived_samples) and isinstance(derived_samples[name], np.ndarray) and not any(np.isnan(derived_samples[name])) and not all(np.array(derived_samples[name])==0):
            ind_good.append(i)
            
    names = [ names[i] for i in ind_good ]
    labels = [ labels[i] for i in ind_good ]
    
    
    ###############################################################################
    #::: if any meaningful values are left, go output them
    ###############################################################################
    if len(names)>0:
            
        #=====================================================================
        #::: save all in pickle
        #=====================================================================
        pickle.dump(derived_samples, open(os.path.join(config.BASEMENT.outdir,mode+'_derived_samples.pickle'),'wb'))
        
        
        #=====================================================================
        #::: save txt & latex table & latex commands
        #=====================================================================
        with open(os.path.join(config.BASEMENT.outdir,mode+'_derived_table.csv'),'w') as outfile,\
             open(os.path.join(config.BASEMENT.outdir,mode+'_derived_latex_table.txt'),'w') as f,\
             open(os.path.join(config.BASEMENT.outdir,mode+'_derived_latex_cmd.txt'),'w') as f_cmd:
                 
            outfile.write('#property,value,lower_error,upper_error,source\n')
            
            f.write('Parameter & Value & Source \\\\ \n')
            f.write('\\hline \n')
            f.write('\\multicolumn{3}{c}{\\textit{Derived parameters}} \\\\ \n')
            f.write('\\hline \n')
            
            for name,label in zip(names, labels):
                ll, median, ul = np.percentile(derived_samples[name], [15.865, 50., 84.135])
                outfile.write( str(label)+','+str(median)+','+str(median-ll)+','+str(ul-median)+',derived\n' )
                
                value = round_tex(median, median-ll, ul-median)
                f.write( label + ' & $' + value + '$ & derived \\\\ \n' )
                
                simplename = name.replace("_", "").replace("/", "over").replace("(", "").replace(")", "").replace("1", "one").replace("2", "two")
                f_cmd.write('\\newcommand{\\'+simplename+'}{$'+value+'$} %'+label+' = $'+value+'$\n')
                
        logprint('\nSaved '+mode+'_derived_results.csv, '+mode+'_derived_latex_table.txt, and '+mode+'_derived_latex_cmd.txt')
        
            
        #=====================================================================
        #::: plot corner
        #=====================================================================
        if 'combined_host_density' in names: names.remove('combined_host_density') #has (N_companions x N_dims) dimensions, thus does not match the rest
        x = np.column_stack([ derived_samples[name] for name in names ])
        fontsize = np.min(( 24. + 0.5*(len(names)), 40 ))
            
        fig = corner(x,
                     range = [0.999]*len(names),
                     labels = names,
                     quantiles=[0.15865, 0.5, 0.84135],
                     show_titles=True, 
                     label_kwargs={"fontsize":fontsize, "rotation":45, "horizontalalignment":'right'},
                     max_n_ticks=3)
        caxes = np.reshape(np.array(fig.axes), (len(names),len(names)))
        
        #::: set allesfitter titles
        for i, name in enumerate(names): 
            
            ll, median, ul = np.percentile(derived_samples[name], [15.865, 50., 84.135])
            value = round_tex(median, median-ll, ul-median)
            ctitle = r'' + labels[i] + '\n' + r'$=' + value + '$'
            if len(names)>1:
                # caxes[i,i].set_title(ctitle)
                caxes[i,i].set_title(ctitle, fontsize=fontsize, rotation=45, horizontalalignment='left')
                for i in range(caxes.shape[0]):
                    for j in range(caxes.shape[1]):
                        caxes[i,j].xaxis.set_label_coords(0.5, -0.5)
                        caxes[i,j].yaxis.set_label_coords(-0.5, 0.5)
            
                        if i==(caxes.shape[0]-1): 
                            fmt = ScalarFormatter(useOffset=False)
                            caxes[i,j].xaxis.set_major_formatter(fmt)
                        if (i>0) and (j==0):
                            fmt = ScalarFormatter(useOffset=False)
                            caxes[i,j].yaxis.set_major_formatter(fmt)
                            
                        for tick in caxes[i,j].xaxis.get_major_ticks(): tick.label.set_fontsize(24) 
                        for tick in caxes[i,j].yaxis.get_major_ticks(): tick.label.set_fontsize(24)    
            else:
                caxes.set_title(ctitle)
                caxes.xaxis.set_label_coords(0.5, -0.5)
                caxes.yaxis.set_label_coords(-0.5, 0.5)
        
        dpi = np.max(( 100. - len(names), 50 ))
        try: #some matplitlib versions cannot handle jpg
            fig.savefig( os.path.join(config.BASEMENT.outdir,mode+'_derived_corner.jpg'), dpi=dpi, bbox_inches='tight' )
        except:
            fig.savefig( os.path.join(config.BASEMENT.outdir,mode+'_derived_corner.png'), bbox_inches='tight' )
        plt.close(fig)
        
        
        #=====================================================================
        #::: finish
        #=====================================================================
        logprint('\nSaved '+mode+'_derived_corner.pdf')
        
        
    else:
        logprint('\nNo values available to be derived.')
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 14:44:29 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt    
from matplotlib.ticker import ScalarFormatter
import os
from shutil import copyfile
import emcee
from corner import corner
import warnings

#::: allesfitter modules
from . import config
from . import deriver
from .general_output import afplot, afplot_per_transit, save_table, save_latex_table, logprint, get_params_from_samples, plot_ttv_results
from .plot_top_down_view import plot_top_down_view
from .utils.latex_printer import round_tex




###############################################################################
#::: draw samples from the MCMC save.5 (internally in the code)
###############################################################################
def draw_mcmc_posterior_samples(sampler, Nsamples=None, as_type='2d_array'):
    '''
    Default: return all possible sampels
    Set e.g. Nsamples=20 for plotting
    '''
#    global config.BASEMENT
    posterior_samples = sampler.get_chain(flat=True, discard=int(1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by']))
    if Nsamples:
        posterior_samples = posterior_samples[np.random.randint(len(posterior_samples), size=Nsamples)]

    if as_type=='2d_array':
        return posterior_samples
    
    elif as_type=='dic':
        posterior_samples_dic = {}
        for key in config.BASEMENT.fitkeys:
            ind = np.where(config.BASEMENT.fitkeys==key)[0]
            posterior_samples_dic[key] = posterior_samples[:,ind].flatten()
        return posterior_samples_dic



###############################################################################
#::: draw the maximum likelihood samples from the MCMC save.5 (internally in the code)
###############################################################################
def draw_mcmc_posterior_samples_at_maximum_likelihood(sampler, as_type='1d_array'):
    log_prob = sampler.get_log_prob(flat=True, discard=int(1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by']))
    posterior_samples = sampler.get_chain(flat=True, discard=int(1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by']))
    ind_max = np.argmax(log_prob)
    posterior_samples = posterior_samples[ind_max,:]
    
    if as_type=='1d_array':
        return posterior_samples

    elif as_type=='dic':
        posterior_samples_dic = {}
        for key in config.BASEMENT.fitkeys:
            ind = np.where(config.BASEMENT.fitkeys==key)[0]
            posterior_samples_dic[key] = posterior_samples[ind].flatten()
        return posterior_samples_dic



###############################################################################
#::: plot the MCMC chains
###############################################################################
def plot_MCMC_chains(sampler):
    
    chain = sampler.get_chain()
    log_prob = sampler.get_log_prob()
    
    #plot chains; emcee_3.0.0 format = (nsteps, nwalkers, nparameters)
    fig, axes = plt.subplots(config.BASEMENT.ndim+1, 1, figsize=(6,3*config.BASEMENT.ndim) )
    
    #::: plot the lnprob_values; emcee_3.0.0 format = (nsteps, nwalkers)
    axes[0].plot(log_prob, '-', rasterized=True)
    axes[0].axvline( 1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by'], color='k', linestyle='--' )
    mini = np.min(log_prob[int(1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by']):,:])
    maxi = np.max(log_prob[int(1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by']):,:])
    axes[0].set( title='lnprob', xlabel='steps', rasterized=True,
                 ylim=[mini, maxi] )
    axes[0].set_xticklabels( [int(label) for label in axes[0].get_xticks()*config.BASEMENT.settings['mcmc_thin_by']] )
    
    #:::plot all chains of parameters
    for i in range(config.BASEMENT.ndim):
        ax = axes[i+1]
        ax.set(title=config.BASEMENT.fitkeys[i], xlabel='steps')
        ax.plot(chain[:,:,i], '-', rasterized=True)
        ax.axvline( 1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by'], color='k', linestyle='--' )
#        ax.set_xticks(ax.get_xticks()[::2])
        ax.set_xticklabels( [int(label) for label in ax.get_xticks()*config.BASEMENT.settings['mcmc_thin_by']] )

    plt.tight_layout()
    return fig, axes
    
    
    
###############################################################################
#::: plot the MCMC corner plot
###############################################################################
#def plot_MCMC_corner(sampler):
#    samples = sampler.get_chain(flat=True, discard=int(1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by']))
#    
#    fig = corner(samples, 
#                 labels = config.BASEMENT.fitkeys,
#                 range = [0.999]*config.BASEMENT.ndim,
#                 quantiles=[0.15865, 0.5, 0.84135],
#                 show_titles=True, title_kwargs={"fontsize": 14},
#                 truths=config.BASEMENT.fittruths)
#            
#    return fig

def plot_MCMC_corner(sampler):
    samples = sampler.get_chain(flat=True, discard=int(1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by']))
    
    params_median, params_ll, params_ul = get_params_from_samples(samples)
    params_median2, params_ll2, params_ul2 = params_median.copy(), params_ll.copy(), params_ul.copy()
    fittruths2 = config.BASEMENT.fittruths.copy()

    #::: make pretty titles for the corner plot  
    labels, units = [], []
    for i,l in enumerate(config.BASEMENT.fitlabels):
        labels.append( str(config.BASEMENT.fitlabels[i]) )
        units.append( str(config.BASEMENT.fitunits[i]) )
    
    for companion in config.BASEMENT.settings['companions_all']:
        if companion+'_epoch' in config.BASEMENT.fitkeys:
            ind = np.where(config.BASEMENT.fitkeys==companion+'_epoch')[0][0]
            samples[:,ind] -= int(params_median[companion+'_epoch'])                #np.round(params_median[companion+'_epoch'],decimals=0)
            units[ind] = str(units[ind]+'-'+str(int(params_median[companion+'_epoch']))+'d')    #np.format_float_positional(params_median[companion+'_epoch'],0)+'d')
            fittruths2[ind] -= int(params_median[companion+'_epoch'])
            params_median2[companion+'_epoch'] -= int(params_median[companion+'_epoch'])
                
    for i,l in enumerate(labels):
        if len( units[i].strip(' ') ) > 0:
            labels[i] = str(labels[i]+' ('+units[i]+')')
        
    #::: corner plot
    fontsize = np.min(( 24. + 0.5*config.BASEMENT.ndim, 40 ))
    fig = corner(samples, 
                 labels = labels,
                 range = [0.999]*config.BASEMENT.ndim,
                 quantiles=[0.15865, 0.5, 0.84135],
                 show_titles=False, 
                 #title_kwargs={"fontsize": 14},
                 label_kwargs={"fontsize":fontsize, "rotation":45, "horizontalalignment":'right'},
                 max_n_ticks=3,
                 truths=fittruths2, truth_color="r")
    caxes = np.reshape(np.array(fig.axes), (config.BASEMENT.ndim,config.BASEMENT.ndim))

    #::: set allesfitter titles
    for i, key in enumerate(config.BASEMENT.fitkeys): 
        
        value = round_tex(params_median2[key], params_ll2[key], params_ul2[key])
        ctitle = r'' + labels[i] + '\n' + r'$=' + value + '$'
        if len(config.BASEMENT.fitkeys)>1:
            # caxes[i,i].set_title(ctitle)
            caxes[i,i].set_title(ctitle, fontsize=fontsize, rotation=45, horizontalalignment='left')
            for i in range(caxes.shape[0]):
                for j in range(caxes.shape[1]):
                    caxes[i,j].xaxis.set_label_coords(0.5, -0.5)
                    caxes[i,j].yaxis.set_label_coords(-0.5, 0.5)
        
                    if i==(caxes.shape[0]-1): 
                        fmt = ScalarFormatter(useOffset=False)
                        caxes[i,j].xaxis.set_major_formatter(fmt)
                    if (i>0) and (j==0):
                        fmt = ScalarFormatter(useOffset=False)
                        caxes[i,j].yaxis.set_major_formatter(fmt)
                        
                    for tick in caxes[i,j].xaxis.get_major_ticks(): tick.label.set_fontsize(24) 
                    for tick in caxes[i,j].yaxis.get_major_ticks(): tick.label.set_fontsize(24)    
        else:
            caxes.set_title(ctitle)
            caxes.xaxis.set_label_coords(0.5, -0.5)
            caxes.yaxis.set_label_coords(-0.5, 0.5)
            
            
    return fig



###############################################################################
#::: print autocorr
###############################################################################
def print_autocorr(sampler):
    logprint('\nConvergence check:')
    logprint('--------------------------')
    
    logprint('{0: <20}'.format('Total steps:'),        '{0: <10}'.format(config.BASEMENT.settings['mcmc_total_steps']))
    logprint('{0: <20}'.format('Burn steps:'),         '{0: <10}'.format(config.BASEMENT.settings['mcmc_burn_steps']))
    logprint('{0: <20}'.format('Evaluation steps:'),   '{0: <20}'.format(config.BASEMENT.settings['mcmc_total_steps'] - config.BASEMENT.settings['mcmc_burn_steps']))
    
    N_evaluation_samples = int( 1. * config.BASEMENT.settings['mcmc_nwalkers'] * (config.BASEMENT.settings['mcmc_total_steps']-config.BASEMENT.settings['mcmc_burn_steps']) / config.BASEMENT.settings['mcmc_thin_by'] )
    logprint('{0: <20}'.format('Evaluation samples:'),   '{0: <20}'.format(N_evaluation_samples))
     
    # if N_evaluation_samples>200000:
    #     answer = input('It seems like you are asking for ' + str(N_evaluation_samples) + 'MCMC evaluation samples (calculated as mcmc_nwalkers * (mcmc_total_steps-mcmc_burn_steps) / mcmc_thin_by).'+\
    #                     'That is an aweful lot of samples.'+\
    #                     'What do you want to do?\n'+\
    #                     '1 : continue at any sacrifice\n'+\
    #                     '2 : abort and increase the mcmc_thin_by parameter in settings.csv (do not do this if you continued an old run!)\n')
    #     if answer==1: 
    #         pass
    #     else:
    #         raise ValueError('User aborted the run.')


    discard=int(1.*config.BASEMENT.settings['mcmc_burn_steps']/config.BASEMENT.settings['mcmc_thin_by'])
    tau = sampler.get_autocorr_time(discard=discard, c=5, tol=10, quiet=True)*config.BASEMENT.settings['mcmc_thin_by']
    logprint('Autocorrelation times:')
    logprint('\t', '{0: <30}'.format('parameter'), '{0: <20}'.format('tau (in steps)'), '{0: <20}'.format('Chain length (in multiples of tau)'))
    converged = True
    for i, key in enumerate(config.BASEMENT.fitkeys):
        chain_length = ((config.BASEMENT.settings['mcmc_total_steps'] - config.BASEMENT.settings['mcmc_burn_steps']) / tau[i])
        logprint('\t', '{0: <30}'.format(key), '{0: <20}'.format(tau[i]), '{0: <20}'.format(chain_length))
        if (chain_length < 30) or np.isinf(chain_length) or np.isnan(chain_length):
            converged = False
            
    if converged:
        logprint('\nSuccesfully converged! All chains are at least 30x the autocorrelation length.\n')
    else:
        logprint('\nNot yet converged! Some chains are less than 30x the autocorrelation length. Please continue to run with longer chains, or start again with more walkers.\n')
        

###############################################################################
#::: analyse the output from save_mcmc.h5 file
###############################################################################
def mcmc_output(datadir, quiet=False):
    '''
    Inputs:
    -------
    datadir : str
        the working directory for allesfitter
        must contain all the data files
        output directories and files will also be created inside datadir
            
    Outputs:
    --------
    This will output information into the console, and create a output files 
    into datadir/results/ (or datadir/QL/ if QL==True)    
    '''
    config.init(datadir, quiet=quiet)
    
    
    #::: security check
    if os.path.exists(os.path.join(config.BASEMENT.outdir,'mcmc_table.csv')):
        try:
            overwrite = str(input('MCMC output files already exists in '+config.BASEMENT.outdir+'.\n'+\
                                  'What do you want to do?\n'+\
                                  '1 : overwrite the output files\n'+\
                                  '2 : abort\n'))
            if (overwrite == '1'):
                pass
            else:
                raise ValueError('User aborted operation.')
        except EOFError:
            warnings.warn("MCMC output files already existed from a previous run, and were automatically overwritten.")
            pass
    
    
    #::: load the mcmc_save.h5
    #::: copy over into tmp file (in case chain is still running and you want a quick look already)     
    copyfile(os.path.join(config.BASEMENT.outdir,'mcmc_save.h5'), os.path.join(config.BASEMENT.outdir,'mcmc_save_tmp.h5'))
    reader = emcee.backends.HDFBackend( os.path.join(config.BASEMENT.outdir,'mcmc_save_tmp.h5'), read_only=True )
    completed_steps = reader.get_chain().shape[0]*config.BASEMENT.settings['mcmc_thin_by']
    if completed_steps < config.BASEMENT.settings['mcmc_total_steps']: 
        #go into quick look mode
        #check how many total steps are actually done so far:
        config.BASEMENT.settings['mcmc_total_steps'] = config.BASEMENT.settings['mcmc_thin_by']*reader.get_chain().shape[0]
        #if this is at least twice the wished-for burn_steps, then let's keep those
        #otherwise, set burn_steps automatically to 75% of how many total steps are actually done so far
        if config.BASEMENT.settings['mcmc_total_steps'] > 2*config.BASEMENT.settings['mcmc_burn_steps']:
            pass
        else:
            config.BASEMENT.settings['mcmc_burn_steps'] = int(0.75*config.BASEMENT.settings['mcmc_total_steps'])
    
    
    #::: print autocorr
    print_autocorr(reader)


    #::: plot the fit
    posterior_samples = draw_mcmc_posterior_samples(reader, Nsamples=20) #only 20 samples for plotting
    
    for companion in config.BASEMENT.settings['companions_all']:
        fig, axes = afplot(posterior_samples, companion)
        if fig is not None:
            fig.savefig( os.path.join(config.BASEMENT.outdir,'mcmc_fit_'+companion+'.pdf'), bbox_inches='tight' )
            plt.close(fig)
        
    for companion in config.BASEMENT.settings['companions_phot']:
        for inst in config.BASEMENT.settings['inst_phot']:
            try:
                fig, axes = afplot_per_transit(posterior_samples, inst, companion)
                fig.savefig( os.path.join(config.BASEMENT.outdir,'mcmc_fit_per_transit_'+inst+'_'+companion+'.pdf'), bbox_inches='tight' )
                plt.close(fig)
            except:
                pass
    
    #::: plot the chains
    fig, axes = plot_MCMC_chains(reader)
    try: #some matplotlib versions cannot handle jpg
        fig.savefig( os.path.join(config.BASEMENT.outdir,'mcmc_chains.jpg'), bbox_inches='tight' )
    except:
        fig.savefig( os.path.join(config.BASEMENT.outdir,'mcmc_chains.png'), bbox_inches='tight' )
    plt.close(fig)


    #::: plot the corner
    fig = plot_MCMC_corner(reader)
    fig.savefig( os.path.join(config.BASEMENT.outdir,'mcmc_corner.pdf'), bbox_inches='tight' )
    plt.close(fig)


    #::: save the tables
    posterior_samples = draw_mcmc_posterior_samples(reader) #all samples
    save_table(posterior_samples, 'mcmc')
    save_latex_table(posterior_samples, 'mcmc')
    
    
    #::: derive values (using stellar parameters from params_star.csv)
    if os.path.exists( os.path.join(config.BASEMENT.datadir,'params_star.csv') ):
    # try:
        deriver.derive(posterior_samples, 'mcmc')
    # except:
        # logprint('\nDerived values could not be produced.')
    else:
        logprint('File "params_star.csv" not found. Cannot derive final parameters.')
    
    
    #::: make top-down orbit plot (using stellar parameters from params_star.csv)
    try:
        params_median, params_ll, params_ul = get_params_from_samples(posterior_samples)
        params_star = np.genfromtxt( os.path.join(config.BASEMENT.datadir,'params_star.csv'), delimiter=',', names=True, dtype=None, encoding='utf-8', comments='#' )
        fig, ax = plot_top_down_view(params_median, params_star)
        fig.savefig( os.path.join(config.BASEMENT.outdir,'top_down_view.pdf'), bbox_inches='tight' )
        plt.close(fig)        
    except:
        logprint('\nOrbital plots could not be produced.')
        
        
    #::: plot TTV results (if wished for)
    if config.BASEMENT.settings['fit_ttvs'] == True:
        plot_ttv_results(params_median, params_ll, params_ul)
        
        
    #::: clean up and delete the tmp file
    os.remove(os.path.join(config.BASEMENT.outdir,'mcmc_save_tmp.h5'))
    
    logprint('\nDone. For all outputs, see', config.BASEMENT.outdir, '\n')
    
    
    #::: return a nerdy quote
    try:
        with open(os.path.join(os.path.dirname(__file__), 'utils', 'quotes.txt')) as dataset:
            return(np.random.choice([l for l in dataset]))
    except:
        return('42')
    
    

###############################################################################
#::: get MCMC samples (for top-level user)
###############################################################################
def get_mcmc_posterior_samples(datadir, Nsamples=None, as_type='dic'): #QL=False, 
    # config.init(datadir, QL=QL)
    config.init(datadir)
    reader = emcee.backends.HDFBackend( os.path.join(config.BASEMENT.outdir,'mcmc_save.h5'), read_only=True )
    return draw_mcmc_posterior_samples(reader, Nsamples=Nsamples, as_type=as_type) #only 20 samples for plotting
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 30 16:58:19 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm
from glob import glob
from pprint import pprint

#::: my modules
from .. import get_mcmc_posterior_samples, get_ns_posterior_samples, get_labels

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




def plot_histograms(datadirs, titles, keys, options=None):
        
    if options is None: options = {}
    if 'type' not in options: options['type'] = 'kde' #or 'hist'
    if 'layout' not in options: options['layout'] = '1x1' #or '5x2' etc.
    if 'simplify' not in options: options['simplify'] = False
    if 'colors' not in options: options['colors'] = None
    if 'epoch_median' not in options: options['epoch_median'] = False
    if 'show_plot' not in options: options['show_plot'] = False
    if 'save_plot' not in options: options['save_plot'] = False
    if 'outdir' not in options: options['outdir'] = '.'
    if 'font_scale' not in options: options['font_scale'] = 1.5 + 0.2*(len(keys)-1)

    
    #::: plot settings (need to renew everything 'cause seaborn likes to go cray cray)
    sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=options['font_scale'], color_codes=True)
    sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
    sns.set_context(rc={'lines.markeredgewidth': 1})
    if options['colors'] is not None: 
        pal = sns.color_palette()
        for i,c in enumerate(options['colors']): pal[i] = c
        sns.set_palette(pal)
            
        
    # alles_list = [ allesclass(x) for x in datadirs ] #only load allesclass objects once to save time
    all_params = {}
    all_paramslabels = {}
    for datadir, title in zip(datadirs, titles):
        try: all_params[title] = get_mcmc_posterior_samples(datadir, as_type='dic')
        except: all_params[title] = get_ns_posterior_samples(datadir, as_type='dic')
        all_paramslabels[title] = get_labels(datadir, as_type='dic')
        
    if options['epoch_median']:
        for title in titles:
            for key in keys:
                if 'epoch' in key:
                    all_params[title][key] -= np.median(all_params[title][key])
                    all_paramslabels[title][key] += 'median'
        
    def plot1(y, ax, typ, label=None):
        if typ=='hist': 
            sns.distplot(y, ax=ax, hist_kws={'linewidth':0, 'alpha':0.5, 'density':True}, label=label, legend=False)
        if typ=='kde': 
            sns.kdeplot(y, ax=ax, shade=True, alpha=1, label=label, legend=False)
            sns.kdeplot(y, ax=ax, shade=False, alpha=1, color='k')
        
    
    if options['layout'] == '1x1':
        fig_list, ax_list = [], []
        for i, key in enumerate(keys):
            fig, ax = plt.subplots()
            # for alles, title in zip(alles_list, titles):
                # plot1(alles.posterior_params[key], ax, options['type'], label=title)
            # ax.set(xlabel=alles.labels[key], ylabel='Posterior density')
            for title in titles:
                plot1(all_params[title][key], ax, options['type'], label=title)
            ax.set(xlabel=all_paramslabels[title][key], ylabel='Posterior density')
            if options['simplify']: setup_simplify(ax)
            fig_list.append(fig)
            ax_list.append(ax)
            if options['show_plot']: plt.show(fig)
            else: plt.close(fig)
            if options['save_plot']: fig.savefig(os.path.join(options['outdir'],'histograms.pdf'), bbox_inches='tight')
        return fig_list, ax_list
        
        
    else:
        fig, axes, rows, cols = setup_layout(options['layout'])
        fig.text(0, 0.5, 'Posterior density', ha='center', va='center', rotation='vertical')
        if rows*cols < len(keys): raise ValueError('Given layout must allow at least as many axes as len(keys).')
        for i, key in enumerate(keys):
            # for alles, title in zip(alles_list, titles):
            #     if rows==1 or cols==1: ax = axes[i]
            #     else: ax = axes[ np.unravel_index(i,(rows,cols)) ]
            #     if key==keys[-1]: plot1(alles.posterior_params[key], ax, options['type'], label=title) #only show one label for the last key
            #     else: plot1(alles.posterior_params[key], ax, options['type'])
            #     ax.set(xlabel=alles.labels[key])
            for title in titles:
                if rows==1 or cols==1: ax = axes[i]
                else: ax = axes[ np.unravel_index(i,(rows,cols)) ]
                plot1(all_params[title][key], ax, options['type'], label=title)
                lgd_handles, lgd_labels = ax.get_legend_handles_labels()
                ax.set(xlabel=all_paramslabels[title][key])
                if options['simplify']: setup_simplify(ax)
        for j in range(i+1,len(keys)+1):
            if rows==1 or cols==1: ax = axes[j]
            else: ax = axes[ np.unravel_index(j,(rows,cols)) ]
            ax.axis('off')
        # lgd = fig.legend(lgd_handles, lgd_labels, loc='upper center', ncol=len(titles), bbox_to_anchor=(0.5,1.02))
        # for legobj in lgd_handles: legobj.set_linewidth(4)
        lgd = ax.legend(lgd_handles, lgd_labels, loc='upper right') #last axis gets the legends
        for line in lgd.get_lines(): line.set_linewidth(10)
        if options['show_plot']: plt.show(fig)
        else: plt.close(fig)
        if options['save_plot']: fig.savefig(os.path.join(options['outdir'],'histograms.pdf'), bbox_inches='tight')
        return fig, axes
    


def setup_simplify(ax):
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.set(yticks=[])



def setup_layout(layout):
    '''
    layout : str
        e.g. '1x1' or '2x3' (rows x cols)
    '''
    rows = int(layout.split('x')[0])
    cols = int(layout.split('x')[1])
    fig, axes = plt.subplots(rows, cols, figsize=(4*cols,4*rows), tight_layout=True)
    return fig, axes, rows, cols


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 14 17:55:39 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
import warnings
from pprint import pprint
from datetime import datetime
from astropy.stats import sigma_clip
from wotan import flatten, slide_clip
from transitleastsquares import transitleastsquares as tls
from transitleastsquares import transit_mask, cleaned_array, catalog_info
from ..exoworlds_rdx.lightcurves.index_transits import index_transits
import time as timer
import contextlib

#::: my modules
try:
    from exoworlds.tess import tessio
except:
    pass
from .fast_slide_clip import fast_slide_clip

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

    

###############################################################################
#::: print to logfile
###############################################################################
def logprint(*text, options=None):
    original = sys.stdout
    try:
        with open(os.path.join(options['outdir'],'logfile.log'), 'a' ) as f:
            sys.stdout = f
            print(*text)
    except OSError:
        pass #For unknown reasons, the combination of open() and os.path.join() does not work on some Windows versions
    sys.stdout = original
    
    
    
###############################################################################
#::: pretty-print to logfile
###############################################################################
def logpprint(*text, options=None):
    original = sys.stdout
    try:
        with open(os.path.join(options['outdir'],'logfile.log'), 'a' ) as f:
            sys.stdout = f
            pprint(*text)
    except OSError:
        pass #For unknown reasons, the combination of open() and os.path.join() does not work on some Windows versions
    sys.stdout = original


    
###############################################################################
#::: apply a mask (if wished so)
###############################################################################
def mask(time, flux, flux_err, period, duration, T0):
    intransit = transit_mask(time, period, duration, T0)
    time = time[~intransit]
    flux = flux[~intransit]
    if flux_err is not None:
        flux_err = flux_err[~intransit]
        time, flux, flux_err = cleaned_array(time, flux, flux_err)
    else:
        time, flux = cleaned_array(time, flux)
    return time, flux, flux_err
    


###############################################################################
#::: apply a mask (if wished so)
###############################################################################
def is_multiple_of(a, b, tolerance=0.05):
    a = np.float(a)
    b = np.float(b) 
    result = a % b
    return (abs(result/b) <= tolerance) or (abs((b-result)/b) <= tolerance)



###############################################################################
#::: get TLS kwargs from TICv8
###############################################################################
def get_tls_kwargs_by_tic(tic_id, sigma=3, tls_kwargs=None, quiet=True):
    #mass comes first, radius comes second in the TLS source code for catalog_info()
    u, M_star, M_star_lerr, M_star_uerr, R_star, R_star_lerr, R_star_uerr = catalog_info(TIC_ID=int(tic_id))
    if not quiet:
        print('TICv8 info:')
        print('Quadratic limb darkening u_0, u_1', u[0], u[1])
        print('Stellar radius', R_star, '+', R_star_lerr, '-', R_star_uerr)
        print('Stellar mass', M_star, '+', M_star_lerr, '-', M_star_uerr)
    if tls_kwargs is None: tls_kwargs = {}
    tls_kwargs['R_star']=float(R_star)
    tls_kwargs['R_star_min']=R_star-sigma*R_star_lerr
    tls_kwargs['R_star_max']=R_star+sigma*R_star_uerr
    tls_kwargs['M_star']=float(M_star)
    tls_kwargs['M_star_min']=M_star-sigma*M_star_lerr
    tls_kwargs['M_star_max']=M_star+sigma*M_star_uerr
    tls_kwargs['u']=u    
    return tls_kwargs



###############################################################################
#::: TLS search on an input lightcurve
###############################################################################
def tls_search(time, flux, flux_err,
               known_transits=None,
               tls_kwargs=None,
               wotan_kwargs=None,
               options=None):
    '''
    Summary:
    -------
    This runs TLS on these data with the given infos
    
    Inputs:
    -------
    time : array of flaot
        time stamps of observations
    flux : array of flaot
        normalized flux
    flux_err : array of flaot
        error of normalized flux
        
    Optional Inputs:
    ----------------
    known_transits : None or dict
        >> can be used to mask known transits before running TLS
        if None
            nothing happens
        if dict 
            if one transit is already known, give for example: 
                known_transits = {'period':[1.3], 'duration':[2.1], 'epoch':[245800.0]}
            if multiple transits are already known, give for example: 
                known_transits = {'name':['b','c'], 'period':[1.3, 21.0], 'duration':[2.1, 4.1], 'epoch':[245800.0, 245801.0]}
            'period' is the period of the known transit(s)
            'duration' is the total duration of the known transit(s), i.e. from first ingress point to last egrees point, in days
            'epoch' is the epoch of the known transit(s)
        
    tls_kwargs : None, str or dict:
        >> can be used to fine-tune the TLS algorithm
        if None
            the default parameters will be chosen (see below)
        if 'default'
            the default parameters will be chosen (see below)
        if dict
            a dictionary with the following keywords is expected; 
            missing keywords will be replaced with default values
            R_star : float
                radius of the star (e.g. median)
                default 1 R_sun (from TLS)
            R_star_min : float
                minimum radius of the star (e.g. 1st percentile)
                default 0.13 R_sun (from TLS)
            R_star_max : float
                maximum radius of the star (e.g. 99th percentile)
                default 3.5 R_sun (from TLS)
            M_star : float
                mass of the star (e.g. median)
                default 1. M_sun (from TLS)
            M_star_min : float
                minimum mass of the star (e.g. 1st percentile)
                default 0.1 M_sun (from TLS)
            M_star_max : float
                maximum mass of the star (e.g. 99th percentile)
                default 1. M_sun (from TLS)    
            u : list
                quadratic limb darkening parameters
                default [0.4804, 0.1867]
            SNR_threshold : float
                the SNR threshold at which to stop the TLS search
                default 5
            SDE_threshold : float
                the SDE threshold at which to stop the TLS search
                default -inf
            FAP_threshold : float
                the False Alarm Probability threshold at which to stop the TLS search
                default inf
        
    wotan_kwargs : None, str, or dict:
        >> can be used to detrend the data before the TLS search
        if None
            the default detrending will run (see below)
        if str is 'default'
            the default detrending will run (see below)
        if str is 'off'
            no detrending will run
        if dict
            a dictionary with two sub-dictionaries is expected; 
            missing keywords will be replaced with default values
            wotan_kwargs['slide_clip'] : dict
                this dictionary contains all slide clipping arguments
                window_length : float
                    slide clip window length
                    default 1
                low : float
                    slide clip lower sigma
                    default 20
                high : float
                    slide clip upper sigma
                    default 3
            wotan_kwargs['flatten'] : dict
                this dictionary contains contains all detrending arguments
                method : str
                    detrending method
                    default 'biweight'
                window_length : float
                    detrending window length in days
                    default 1         
            
    options : None or dict, keywords:
        >> can be used for any general options
        if None
            the default options will be used (see below)
        if dict
            a dcitionary with the following keywords is expected;
            missing keywords will be replaced with default values
            show_plot : bool
                can show a plot of each phase-folded transit candidate and TLS model in the terminal 
                default is True
            save_plot : bool or str
                can save a plot of each phase-folded transit candidate and TLS model into outdir
                if True, will be set to '123'
                if str, then: '1': detrended plot, '2': TLS plot, '3': all TLS plots, and any combinations thereof
                default is True
            save_csv : bool
                can save a csv of the detrended lightcurve
                default is True
            outdir : string
                if None
                    a new directory called "results" will be created in the current folder
                default is "tls_results_[wotan_flatten_method]_[wotan_flatten_window_length]"
        
    Returns:
    -------
    List of all TLS results
    '''
    
    #::: seeed
    np.random.seed(42)
    
    
    #::: handle inputs
    if flux_err is None:
        ind = np.where( ~np.isnan(time*flux) )[0]
        time = time[ind]
        flux = flux[ind]
    else:
        ind = np.where( ~np.isnan(time*flux*flux_err) )[0]
        time = time[ind]
        flux = flux[ind]
        flux_err = flux_err[ind]
    
    time_input = 1.*time
    flux_input = 1.*flux #for plotting
        
    
    if type(wotan_kwargs)==str and wotan_kwargs=='off': 
        detrend = False
    else:
        detrend = True
        if (wotan_kwargs is None) or (type(wotan_kwargs)==str and wotan_kwargs=='default'): wotan_kwargs={} 
        if 'slide_clip' not in wotan_kwargs: wotan_kwargs['slide_clip'] = {}
        if wotan_kwargs['slide_clip'] is not None:
            if 'window_length' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['window_length'] = 1.
            if 'low' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['low'] = 20.
            if 'high' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['high'] = 3.
    
    if 'flatten' not in wotan_kwargs: wotan_kwargs['flatten'] = {}
    if wotan_kwargs['flatten'] is not None:
        if 'method' not in wotan_kwargs['flatten']: wotan_kwargs['flatten']['method'] = 'biweight'
        if 'window_length' not in wotan_kwargs['flatten']: wotan_kwargs['flatten']['window_length'] = 1.
    #the rest is filled automatically by Wotan
        
    if tls_kwargs is None: tls_kwargs = {}
    if 'show_progress_bar' not in tls_kwargs: tls_kwargs['show_progress_bar'] = False
    if 'SNR_threshold' not in tls_kwargs: tls_kwargs['SNR_threshold'] = 5.
    if 'SDE_threshold' not in tls_kwargs: tls_kwargs['SDE_threshold'] = -np.inf #don't trust SDE
    if 'FAP_threshold' not in tls_kwargs: tls_kwargs['FAP_threshold'] = np.inf #don't trust FAP 
    tls_kwargs_original = {key: tls_kwargs[key] for key in tls_kwargs.keys() if key not in ['SNR_threshold','SDE_threshold','FAP_threshold']} #for the original tls
    #the rest is filled automatically by TLS
    
    if options is None: options = {}
    if 'show_plot' not in options: options['show_plot'] = True
    if type(options['show_plot'])==bool and (options['show_plot'] is True): options['show_plot']='123' #1: detrended plot, 2: TLS plot, 3: all TLS plots
    if type(options['show_plot'])==bool and (options['show_plot'] is False): options['show_plot']='' #1: detrended plot, 2: TLS plot, 3: all TLS plots
    if 'save_plot' not in options: options['save_plot'] = True
    if type(options['save_plot'])==bool and (options['save_plot'] is True): options['save_plot']='123' #1: detrended plot, 2: TLS plot, 3: all TLS plots
    if type(options['save_plot'])==bool and (options['save_plot'] is False): options['save_plot']='' #1: detrended plot, 2: TLS plot, 3: all TLS plots
    if 'save_csv' not in options: options['save_csv'] = True
    if 'outdir' not in options: 
        if detrend:
            options['outdir'] = 'tls_results_'+wotan_kwargs['flatten']['method']+'_'+str(wotan_kwargs['flatten']['window_length'])
        else:
            options['outdir'] = 'tls_results_undetrended'
    if 'quiet' not in options: options['quiet'] = True
    if 'inj_period' not in options: options['inj_period'] = np.nan
    
    
    #::: init
    SNR = 1e12
    SDE = 1e12
    FAP = 0
    FOUND_SIGNAL = False
    results_all = []      
    if len(options['outdir'])>0 and not os.path.exists(options['outdir']): os.makedirs(options['outdir'])  
    
    
    #::: logprint
    with open( os.path.join(options['outdir'], 'logfile.log'), 'w' ) as f:
        f.write('TLS search, UTC ' + datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S') + '\n')
    logprint('\nWotan kwargs:', options=options)
    logpprint(wotan_kwargs, options=options)
    logprint('\nTLS kwargs:', options=options)
    logpprint(tls_kwargs, options=options)
    logprint('\nOptions:', options=options)
    logpprint(options, options=options)
    
    # timer1 = timer.time()
    # print('t1', timer1 - timer0)
            
    #::: apply a mask (if wished so)
    if known_transits is not None:
        for period, duration, T0 in zip(known_transits['period'], known_transits['duration'], known_transits['epoch']):
            time, flux, flux_err = mask(time, flux, flux_err, period, duration, T0)
    
    
    #::: global sigma clipping
    flux = sigma_clip(flux, sigma_upper=3, sigma_lower=20)
    
    # timer2 = timer.time()
    # print('t2', timer2 - timer0)
    
    #::: detrend (if wished so)
    if detrend:
        
        #::: slide clipping (super slow)
        # if wotan_kwargs['slide_clip'] is not None: flux = slide_clip(time, flux, **wotan_kwargs['slide_clip']) #slide_clip is super slow (10 seconds for a TESS 2 min lightcurve for a single Sector)
        # timer3a = timer.time()
        # print('t3a', timer3a - timer0)   
    
        #::: fast slide clipping (super fast)
        if wotan_kwargs['slide_clip'] is not None: flux = fast_slide_clip(time, flux, **wotan_kwargs['slide_clip']) #slide_clip is super fast (<1 seconds for a TESS 2 min lightcurve for a single Sector)
        # timer3a = timer.time()
        # print('t3a', timer3a - timer0)   
        
        #::: detrending (super fast)
        if wotan_kwargs['flatten'] is not None: flux, trend = flatten(time, flux, return_trend=True, **wotan_kwargs['flatten']) #flatten is super fast, (<1 second for a TESS 2 min lightcurve for a single Sector)
        # timer3b = timer.time()
        # print('t3b', timer3b - timer0)   
        
        #::: global sigma clipping on the flattened flux (super fast)
        flux = sigma_clip(flux, sigma_upper=3, sigma_lower=20)
        # timer3c = timer.time()
        # print('t3c', timer3c - timer0)   
        
        if ('1' in options['show_plot']) or ('1' in options['save_plot']):
            fig, axes = plt.subplots(2,1, figsize=(40,8))
            axes[0].plot(time, flux_input, 'b.', rasterized=True)
            axes[0].plot(time, trend, 'r-', lw=2)
            axes[0].set(ylabel='Flux (input)', xticklabels=[])
            axes[1].plot(time, flux, 'b.', rasterized=True)
            axes[1].set(ylabel='Flux (detrended)', xlabel='Time (BJD)')
        if ('1' in options['save_plot']):
            try: fig.savefig(os.path.join(options['outdir'],'flux_'+wotan_kwargs['flatten']['method']+'.pdf'), bbox_inches='tight') #some matplotlib versions crash when saving pdf...
            except: fig.savefig(os.path.join(options['outdir'],'flux_'+wotan_kwargs['flatten']['method']+'.jpg'), bbox_inches='tight') #some matplotlib versions need pillow for jpg (conda install pillow)...
            if ('1' in options['show_plot']):
                plt.show(fig)
            else:
                plt.close(fig)
                
        if options['save_csv']:
            if flux_err is None: flux_err0 = np.nan*time
            else: flux_err0 = flux_err
            X = np.column_stack((time, flux, flux_err0, trend))
            np.savetxt(os.path.join(options['outdir'],'flux_'+wotan_kwargs['flatten']['method']+'.csv'), X, delimiter=',', header='time,flux_detrended,flux_err,trend')
        
        time_detrended = 1.*time #just for plotting
        flux_detrended = 1.*flux #just for plotting
        
    # timer3d = timer.time()
    # print('t3d', timer3d - timer0)    
    
    
    #::: search for transits
    i = 0
    ind_trs = []
    while (SNR >= tls_kwargs['SNR_threshold']) and (SDE >= tls_kwargs['SDE_threshold']) and (FAP <= tls_kwargs['FAP_threshold']) and (FOUND_SIGNAL==False):
        
        if options['quiet']:
            with open(os.devnull, 'w') as devnull:
                with contextlib.redirect_stdout(devnull):
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        model = tls(time, flux, flux_err)
                        results = model.power(**tls_kwargs_original)
        else:
            model = tls(time, flux, flux_err)
            results = model.power(**tls_kwargs_original)
            
        # timer4 = timer.time()
        # print('t4', timer4 - timer0)  
        
        # plt.figure()
        # plt.plot(time, flux, 'b.')
        # pprint(tls_kwargs_original)
        # pprint(results)
        # err
        
        if (results.snr >= tls_kwargs['SNR_threshold']) and (results.SDE >= tls_kwargs['SDE_threshold']) and (results.FAP <= tls_kwargs['FAP_threshold']):
            
            #::: calculcate the correct_duration, as TLS sometimes returns unreasonable durations
            ind_tr_phase = np.where( results['model_folded_model'] < 1. )[0]
            correct_duration = results['period'] * (results['model_folded_phase'][ind_tr_phase[-1]] - results['model_folded_phase'][ind_tr_phase[0]])
            
            #::: mark transit
            ind_tr, ind_out = index_transits(time_input, results['T0'], results['period'], correct_duration)
            ind_trs.append(ind_tr)
            
            #::: mask out detected transits and append results
            time1, flux1 = time, flux #for plotting
            time, flux, flux_err = mask(time, flux, flux_err, results.period, np.max((1.5*correct_duration)), results.T0)
            results_all.append(results)
            
            #::: write TLS stats to file
            with open(os.path.join(options['outdir'],'tls_signal_'+str(i)+'.txt'), 'wt') as out:
                pprint(results, stream=out)
    
            # timer5 = timer.time()
            # print('t5', timer5 - timer0)  
    
            #::: individual TLS plots
            if ('2' in options['show_plot']) or ('2' in options['save_plot']):
                fig = plt.figure(figsize=(20,8), tight_layout=True)
                gs = fig.add_gridspec(2,3)
                
                ax = fig.add_subplot(gs[0,:])
                ax.plot(time1, flux1, 'b.', rasterized=True)
                ax.plot(results['model_lightcurve_time'], results['model_lightcurve_model'], 'r-', lw=3)
                ax.set(xlabel='Time (BJD)', ylabel='Flux')
                
                ax = fig.add_subplot(gs[1,0])
                ax.plot(results['folded_phase'], results['folded_y'], 'b.', rasterized=True)
                ax.plot(results['model_folded_phase'], results['model_folded_model'], 'r-', lw=3)
                ax.set(xlabel='Phase', ylabel='Flux')
                
                ax = fig.add_subplot(gs[1,1])
                ax.plot((results['folded_phase']-0.5)*results['period']*24, results['folded_y'], 'b.', rasterized=True)
                ax.plot((results['model_folded_phase']-0.5)*results['period']*24, results['model_folded_model'], 'r-', lw=3)
                ax.set(xlim=[ -1.5*correct_duration*24, +1.5*correct_duration*24 ], xlabel='Time (h)', yticks=[])
                
                ax = fig.add_subplot(gs[1,2])
                ax.text( .02, 0.95, 'P = ' + np.format_float_positional(results['period'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
                ax.text( .02, 0.85, 'Depth = ' + np.format_float_positional(1e3*(1.-results['depth']),4) + ' ppt', ha='left', va='center', transform=ax.transAxes )
                ax.text( .02, 0.75, 'Duration = ' + np.format_float_positional(24*correct_duration,4) + ' h', ha='left', va='center', transform=ax.transAxes )
                ax.text( .02, 0.65, 'T_0 = ' + np.format_float_positional(results['T0'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
                ax.text( .02, 0.55, 'SNR = ' + np.format_float_positional(results['snr'],4), ha='left', va='center', transform=ax.transAxes )
                ax.text( .02, 0.45, 'SDE = ' + np.format_float_positional(results['SDE'],4), ha='left', va='center', transform=ax.transAxes )
                ax.text( .02, 0.35, 'FAP = ' + np.format_float_scientific(results['FAP'],4), ha='left', va='center', transform=ax.transAxes )
                ax.set_axis_off()
                if ('2' in options['save_plot']):
                    try: fig.savefig(os.path.join(options['outdir'],'tls_signal_'+str(i)+'.pdf'), bbox_inches='tight') #some matplotlib versions crash when saving pdf...
                    except: fig.savefig(os.path.join(options['outdir'],'tls_signal_'+str(i)+'.jpg'), bbox_inches='tight') #some matplotlib versions need pillow for jpg (conda install pillow)...
                if ('2' in options['show_plot']):
                    plt.show(fig)
                else:
                    plt.close(fig)
                    
            # timer6 = timer.time()
            # print('t6', timer6 - timer0)  
            
        SNR = results.snr
        SDE = results.SDE
        FAP = results.FAP
        if is_multiple_of(results['period'],options['inj_period']): SNR = -np.inf #if run as part of an inejction-recovery test, then abort if it matches the injected period
        i+=1
        
        
        
    #::: full lightcurve plot
    if ('3' in options['show_plot']) or ('3' in options['save_plot']):
        
        if detrend:
            fig, axes = plt.subplots(2,1, figsize=(40,8), tight_layout=True)
            ax = axes[0]
            ax.plot(time_input, flux_input, 'k.', color='grey', rasterized=True)
            ax.plot(time_input, trend, 'r-', lw=2)
            for number, ind_tr in enumerate(ind_trs):
                ax.plot(time_input[ind_tr], flux_input[ind_tr], marker='.', linestyle='none', label='signal '+str(number))
            ax.set(ylabel='Flux (input)', xticklabels=[])
            ax.legend()

            ax = axes[1]
            ax.plot(time_detrended, flux_detrended, 'k.', color='grey', rasterized=True)
            for number, ind_tr in enumerate(ind_trs):
                ax.plot(time_detrended[ind_tr], flux_detrended[ind_tr], marker='.', linestyle='none', label='signal '+str(number))
            ax.set(ylabel='Flux (detrended)', xlabel='Time (BJD)')
            ax.legend()
            
        else:
            fig = plt.figure(figsize=(20,4), tight_layout=True)
            fig, ax = plt.subplots(1,1, figsize=(40,4))
            ax.plot(time_input, flux_input, 'k.', color='grey', rasterized=True)
            ax.set(ylabel='Flux (input)', xlabel='Time (BJD)')
            for number, ind_tr in enumerate(ind_trs):
                ax.plot(time_input[ind_tr], flux_input[ind_tr], marker='.', linestyle='none', label='signal '+str(number))
            ax.legend()
        
        if ('3' in options['save_plot']):
            try: fig.savefig(os.path.join(options['outdir'],'tls_signal_all.pdf'), bbox_inches='tight') #some matplotlib versions crash when saving pdf...
            except: fig.savefig(os.path.join(options['outdir'],'tls_signal_all.jpg'), bbox_inches='tight') #some matplotlib versions need pillow for jpg (conda install pillow)...
        if ('3' in options['show_plot']):
            plt.show(fig)
        else:
            plt.close(fig)                    
                
            
    return results_all



###############################################################################
#::: TLS search using tessio
###############################################################################
def tls_search_by_tic(tic_id,
                      tls_kwargs=None, SNR_threshold=5., known_transits=None,
                      options=None):
    '''
    Summary:
    -------
    wrapper around tls_search()
    retrieves the SPOC PDC-SAP lightcurve
    retrieves all TIC catalog information from MAST
    calls tls_search()
    
    Inputs:
    -------
    tic_id : str
        TIC ID
        
    Optional Inputs:
    ----------------
    see tls_search()
        
    Returns:
    -------
    list of all TLS results
    '''
    
    #::: handle inputs
    if options is None: options = {}
    if 'show_plot' not in options: options['show_plot']=False
    if 'save_plot' not in options: options['save_plot']=False
    if 'outdir' not in options: options['outdir']=''
    
    #::: format inputs
    tic_id = str(int(tic_id))
    
    #::: load data and inject transit
    time, flux, flux_err = tessio.get(tic_id, pipeline='spoc', PDC=True, unpack=True)
    
    #::: load TIC info / tls kwargs
    tls_kwargs = get_tls_kwargs_by_tic(tic_id, tls_kwargs=tls_kwargs)
    
    return tls_search(time, flux, flux_err,
                      tls_kwargs=tls_kwargs,
                      SNR_threshold=SNR_threshold,
                      known_transits=known_transits,
                      options=options)



###############################################################################
#::: main
###############################################################################
if __name__ == '__main__':
    pass
    
    ###########################################################################
    #::: Example: search for a transit with TLS and tessio
    ###########################################################################
    # tic_id = '269701147'
    # SNR_threshold=5.,
    # known_transits = {'epoch':[2458715.3547, 2458726.0526, 2458743.5534],
    #                   'period':[8.8806, 28.5810, 38.3497],
    #                   'duration':[3.09/24., 4.45/24., 5.52/24.]
    #                  }
    
    # results_all = tls_search_by_tic(tic_id,
    #                 SNR_threshold=SNR_threshold,
    #                 known_transits=known_transits)
    # print(results_all)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 21 16:49:37 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from pprint import pprint

#::: my modules
from classes import allesclass2
from translator import translate

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})


#::: user settings
rr = 0.1
rsuma = 0.1
epoch = 0
period = 1
cosi = 0
r_host = 1
m_host = 1
time = np.linspace(-0.5,0.5,1001)

#::: allesclass2
alles = allesclass2()
alles.settings = {'companions_phot':['b'], 'inst_phot':['buffer']}
alles.params = {'b_rr':rr, 'b_rsuma':rsuma, 'b_epoch':epoch, 'b_period':period, 'b_cosi':cosi}
alles.params_host = {'R_host':r_host, 'M_host':m_host}
alles.fill()   

#::: model flux
model_flux = alles.generate_model('buffer', 'flux', time)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  8 18:41:19 2019

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os
from tqdm import tqdm
import itertools
from transitleastsquares import catalog_info

#::: my modules
from .transit_search import tls_search
from ..generator import inject_lc_model
from .injection_recovery_output import irplot
try:
    from exoworlds.tess import tessio
except:
    pass

#::: settings and constants
np.random.seed(42)
eps = 1e-12




###############################################################################
#::: helper functiomns
###############################################################################

#::: set up the logfile
def setup_logfile(logfname):
    '''
    Summary:
    --------
    - sets up a logfile for the injection-recovery tests
    - starts user dialog if a logfile already exists
    
    Inputs:
    -------
    logfname : str
        file path and name for the logfile
    
    Returns:
    -------
    ex : np struct array / None
        existing logfile
    '''
    if os.path.exists(logfname):
        response = input('Log file already exists. Do you want to (1) overwrite it, (2) append missing rows, or (3) abort?\n')
        if response == '1':   
            with open(logfname,'w') as f: 
                f.write('inj_period,inj_rplanet,tls_period,tls_depth,tls_duration,tls_epoch,tls_SDE,tls_SNR\n')
            return None
        elif response == '2': 
            ex = np.genfromtxt(logfname, names=True, dtype=None, delimiter=',')
            return ex
        else:
            raise ValueError('User aborted.') 
    else:
        with open(logfname,'w') as f: 
            f.write('inj_period,inj_rplanet,tls_period,tls_depth,tls_duration,tls_SDE,tls_SNR\n')
        return None                
        


#::: check if a certain injection has already been logged, or if it needs to be done
def to_do_or_not_to_do_that_is_the_question(ex, period, rplanet):
    #if no previous logfile exists, then do it
    if ex is None:
        return True #do it
    #if a previous logfile exists and it includes those period and rplanet values (to 1e-6 precision), then skip this pair
    else:
        for i in range(len(ex)):
            if (np.abs(period - ex['inj_period'][i]) < 1e-6) and (np.abs(rplanet - ex['inj_rplanet'][i]) < 1e-6):
                return False #skip it
    #if the pair was not skipped before, then go ahead and do it
    return True #do it
    


###############################################################################
#::: Inject a grid of ellc transits and TLS search on each lightcurve
###############################################################################
def inject_and_tls_search(time, flux, flux_err,                                #lightcurve
                          periods=1., rplanets=11., logfname='logfile.csv',    #for inject_lightcurve_model()
                          # SNR_threshold=5.,                                  #for tls_search()
                          # known_transits=None, 
                          # mask_multiplier=1.5,
                          # R_host=1., R_host_min=0.13, R_host_max=3.5, 
                          # M_host=1., M_host_min=0.1, M_host_max=1.,
                          # ldc=[0.4804, 0.1867],
                          # n_transits_min=3, 
                          # show_plot=False, save_plot=False, outdir='',
                          injection_params=None,
                          injection_settings=None,
                          tls_settings=None,
                          **kwargs):
    '''
    Summary:
    --------
    - injects a planet signal via ellc, for a given period and radius (at random epoch)
    - runs TLS on these injected data and infos
    
    Inputs:
    -------
    time : array of flaot
        time stamps of observations
    flux : array of flaot
        normalized flux
    flux_err : array of flaot
        error of normalized flux
        
    Optional Inputs:
    ----------------
    **kwargs : keyword arguments
        any other keyword arguments, see inject_lightcurve_model() and tls_search()
    
    Returns:
    -------
    Nothing, but a list of all TLS results will get saved to a log file
    '''

    #::: convert input
    periods = np.atleast_1d(periods)
    rplanets = np.atleast_1d(rplanets)
    if injection_params is None: injection_params = defaults.get_default_params()
    else: injection_params = defaults.fill_params()
        
        
    #::: set up a logfile
    ex = setup_logfile(logfname)
    
    #::: cycle through all period and R_companion
    print('\n', flush=True)
    for period, rplanet in tqdm(itertools.product(periods, rplanets), total=len(periods)*len(rplanets)): #combining the two for loops
        
        #::: check if this was already done
        if to_do_or_not_to_do_that_is_the_question(ex, period, rplanet):
            print('\tP = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth --> do')
            epoch = time[0] + np.random.random()*period
            
            #::: inject
            flux2 = inject_lc_model(time, flux, flux_err, **kwargs)
            
            #::: recover
            results_all = tls_search(time, flux2, flux_err,
                                     period_min=0.5*period, period_max=2*period, #NEW NEW NEW MUCH SPEED MUCH FAST
                                     transit_depth_min=0.5*(1.*rplanet/R_host)**2, #NEW NEW NEW MUCH SPEED MUCH FAST
                                     **kwargs)
            
            #::: check if something was found
            if len(results_all)>0:
                for r in results_all:
                    with open(logfname,'a') as f:
                        f.write(format(period, '.5f')+','+
                                format(rplanet, '.5f')+','+
                                format(r.period, '.5f')+','+
                                format(r.depth, '.5f')+','+
                                format(r.duration, '.5f')+','+
                                format(r.epoch, '.5f')+','+
                                format(r.SDE, '.5f')+','+
                                format(r.snr, '.5f')+'\n')
            else:
                with open(logfname,'a') as f:
                    f.write(format(period, '.5f')+','+
                            format(rplanet, '.5f')+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+'\n')
                
        else:
            print('\tP = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth --> skipped (already exists)')
    
    #::: finish
    irplot(logfname)
    print('Done.')
            



###############################################################################
#::: Inject an ellc transit and TLS search using tessio
###############################################################################
def inject_and_tls_search_by_tic(tic_id, sigma_multiplier=3, **kwargs):
    '''
    Summary:
    -------
    - wrapper around inject_and_tls_search()
    - retrieves the SPOC PDC-SAP lightcurve
    - retrieves all TIC catalog information from MAST
    - injects a planet signal via ellc, for a given period and radius (at random epoch)
    - runs TLS on these injected data and infos
    
    Inputs:
    -------
    tic_id : str or int
        TIC ID
        
    Optional Inputs:
    ----------------
    sigma_multiplier : float 
        set TLS' uniform bounds for the host's radius and mass at sigma_multiplier times the error bars from TICv8
        default is 3
    **kwargs : keyword arguments
        any other keyword arguments, see inject_and_tls_search()
        
    Returns:
    -------
    a list of all TLS results will get saved to a log file
    '''
    
    #::: format inputs
    tic_id = str(int(tic_id))

    #::: load data
    time, flux, flux_err = tessio.get(tic_id, pipeline='spoc', PDC=True, unpack=True)
    
    #::: load TIC info
    ldc, R_host, R_host_lerr, R_host_uerr, M_host, M_host_lerr, M_host_uerr = catalog_info(TIC_ID=int(tic_id))
    print('\nTICv8 info:')
    print('Quadratic limb darkening u_0, u_1', ldc[0], ldc[1])
    print('Stellar radius', R_host, '+', R_host_lerr, '-', R_host_uerr)
    print('Stellar mass', M_host, '+', M_host_lerr, '-', M_host_uerr)
    
    #::: run
    inject_and_tls_search(time, flux, flux_err,
                          R_host=R_host, R_host_min=R_host-sigma_multiplier*R_host_lerr, R_host_max=R_host+sigma_multiplier*R_host_uerr, 
                          M_host=M_host, M_host_min=M_host-sigma_multiplier*M_host_lerr, M_host_max=M_host+sigma_multiplier*M_host_uerr,
                          ldc=ldc,
                          **kwargs)
            


###############################################################################
#::: diagnostic TLS plots
###############################################################################
def tls_plot(results):
    
    fig, axes = plt.subplots(1,2,figsize=(20,5))
    
    ax = axes[0]
    ax.axvline(results.period, alpha=0.4, lw=3)
    ax.set_xlim(np.min(results.period), np.max(results.period))
    for n in range(2, 10):
        ax.axvline(n*results.period, alpha=0.4, lw=1, linestyle="dashed")
        ax.axvline(results.period / n, alpha=0.4, lw=1, linestyle="dashed")
    ax.set_ylabel(r'SDE')
    ax.set_xlabel('Period (days)')
    ax.plot(results.period, results.power, color='k', lw=0.5)
    ax.set_title('Period '+format(results.period, '.5f') + ', SNR '+format(results.snr, '.1f'))
    ax.set_xlim(0, max(results.period))
#    plt.show(fig)
#    fig.savefig('plots/'+'Periodogram P = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth'+'.pdf', bbox_inches='tight')
#    plt.close(fig)

    ax = axes[1]
    ax.plot(results.model_folded_phase, results.model_folded_model, color='red')
    ax.scatter(results.folded_phase, results.folded_y, color='b', s=10, alpha=0.5, zorder=2)
    ax.set_xlim(0.4, 0.6)
    ax.set_ylim(0.99,1.01)
    ax.set_xlabel('Phase')
    ax.set_ylabel('Relative flux');
    ax.set_title('Period '+format(results.period, '.5f'))
#    plt.show(fig)
#    fig.savefig('plots/'+'TLS transit P = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth'+'.pdf', bbox_inches='tight')
#    plt.close(fig)
    
    
    
###############################################################################
#::: main
###############################################################################
if __name__ == '__main__':
    pass
    
    ###########################################################################
    #::: Example: injection recovery test with TLS and tessio
    ###########################################################################
    # tic_id = '269701147'
    # period = np.arange(2,160+eps,2) #in days
    # R_companion = np.arange(0.8,4+eps,0.1) #in R_earth
    # SNR_threshold = 3.
    # known_transits = {'epoch':[2458715.3547, 2458726.0526, 2458743.5534],
    #                   'period':[8.8806, 28.5810, 38.3497],
    #                   'duration':[3.09/24., 4.45/24., 5.52/24.]
    #                   }
    # logfname = 'TIC_269701147.csv'
    
    # inject_and_tls_search_by_tic(tic_id, period, R_companion, logfname, 
    #                               SNR_threshold=SNR_threshold,
    #                               known_transits=known_transits)
    
    

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 31 15:50:31 2020

@author: 
Dr. Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
GitHub: https://github.com/MNGuenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os
import warnings
try:
    from wotan import flatten
except:
    warnings.warn('Install wotan for full TTV preparation functionality.')
    pass

#::: my modules
import allesfitter
from allesfitter.lightcurves.lightcurves import eclipse_width_smart
from allesfitter.exoworlds_rdx.lightcurves.index_transits import get_tmid_observed_transits
from allesfitter.general_output import logprint


    

###############################################################################
#::: prepare TTV fit (if chosen)
###############################################################################
def prepare_ttv_fit(datadir, ax=None):
    '''
    this must be run *after* reduce_phot_data()
    '''
    ax0 = ax
    
    alles = allesfitter.allesclass(datadir)
    window = alles.settings['fast_fit_width']
       
    if not os.path.exists( os.path.join(datadir,'ttv_preparation') ): os.makedirs(os.path.join(datadir,'ttv_preparation'))
    with open(os.path.join(datadir,'ttv_preparation','ttv_initial_guess_params.csv'),'w') as f:
        f.write('')
    
    def plot_all_transits_color_coded():
        for inst in alles.settings['inst_phot']:
            time = alles.data[inst]['time']
            for companion in alles.settings['companions_phot']:
                ind = []
                for i, t in enumerate(alles.data[companion+'_tmid_observed_transits']):
                    ind += list( np.where((time >= (t - window/2.)) & (time <= (t + window/2.)))[0] )
                ax.plot( alles.data[inst]['time'][ind], alles.data[inst]['flux'][ind], ls='none', marker='.', label=companion )
    
    
    for companion in alles.settings['companions_phot']:
        with open(os.path.join(datadir,'ttv_preparation','ttv_initial_guess_params.csv'),'a') as f:
            f.write('#TTV companion '+companion+',,,,,\n')
        
        #----------------------------------------------------------------------
        #::: get combined data from all instruments
        #----------------------------------------------------------------------
        all_times = []
        all_flux = []
        for inst in alles.settings['inst_phot']:
            all_times += list(alles.data[inst]['time'])
            all_flux += list(alles.data[inst]['flux'])
        ind_sort = np.argsort(all_times)
        all_times = np.array(all_times)[ind_sort]
        all_flux = np.array(all_flux)[ind_sort]
        
        #----------------------------------------------------------------------
        #::: get eclipse window
        #----------------------------------------------------------------------
        alles.initial_guess_params_median[companion+'_epoch']
        eclipse_width = eclipse_width_smart(alles.initial_guess_params_median[companion+'_period'], 
                                    alles.initial_guess_params_median[companion+'_rr'], 
                                    alles.initial_guess_params_median[companion+'_rsuma'], 
                                    alles.initial_guess_params_median[companion+'_cosi'], 
                                    alles.initial_guess_params_median[companion+'_f_s'], 
                                    alles.initial_guess_params_median[companion+'_f_c'], 
                                    )[0]
        
        #----------------------------------------------------------------------
        #::: compute tmid, ttv_guess and make per-transit-plots
        #----------------------------------------------------------------------
        tmids = []
        
        alles.data[companion+'_tmid_observed_transits'] = get_tmid_observed_transits(all_times,alles.initial_guess_params_median[companion+'_epoch'],alles.initial_guess_params_median[companion+'_period'],alles.settings['fast_fit_width'])
        N = len(alles.data[companion+'_tmid_observed_transits'])
        fig, axes = plt.subplots(N, 1, figsize=(6,4*N), sharey=True, tight_layout=True)
        
        for i, t in enumerate(alles.data[companion+'_tmid_observed_transits']):
            ind_tr1 = np.where((all_times >= (t - window/2.)) & (all_times <= (t + window/2.)))[0]
            tr_times = all_times[ind_tr1]
            tr_flux = all_flux[ind_tr1]
            t_exp = np.median(np.diff(tr_times))
            N_points_in_eclipse = int(eclipse_width/t_exp)
            try:
                trend = flatten(tr_times, tr_flux, window_length=eclipse_width/2., method='biweight', return_trend=True)[1]
                tmid = np.median( tr_times[ np.argsort(trend)[0:int(N_points_in_eclipse/2.)] ] )
            except:
                trend = None
                tmid = np.median( tr_times[ np.argsort(tr_times)[0:int(N_points_in_eclipse/2.)] ] )
            ttv_guess = tmid - t
            tmids.append(tmid)
            
            ax = axes[i]
            ax.plot(tr_times, tr_flux, 'b.', rasterized=True)
            if trend is not None: ax.plot(tr_times, trend, 'r-')
            ax.axvline(t,c='grey',ls='--',label='linear prediction')
            ax.axvline(tmid,c='r',ls='--',label='flux minimum')
            ax.set(xlabel='Time', ylabel='Flux', xlim=[t-window/2., t+window/2.])
            ax.text(0.95,0.95,'Transit '+str(i+1), va='top', ha='right', transform=ax.transAxes)
            with open(os.path.join(datadir,'ttv_preparation','ttv_initial_guess_params.csv'),'a') as f:
               f.write(companion+'_ttv_transit_'+str(i+1)+','+np.format_float_positional(ttv_guess,4)+',1,uniform '+np.format_float_positional(ttv_guess-0.01,4)+' '+np.format_float_positional(ttv_guess+0.01,4)+',TTV$_\mathrm{'+companion+';'+str(i+1)+'}$,d\n')
        axes[0].legend()
        fig.savefig(os.path.join(datadir,'ttv_preparation','ttv_preparation_'+companion+'_per_transit.pdf'), bbox_inches='tight')
        plt.close(fig)
         
        tmids = np.array(tmids)
        
        
        #----------------------------------------------------------------------
        #::: ttv guess 0-C plot
        #----------------------------------------------------------------------
        nr = np.array([ int(np.round( (t-tmids[0]) / alles.initial_guess_params_median[companion+'_period'] )) for t in tmids ]) #get corresponding transit number
        nr -= int(nr[-1]/2.) #shift into the middle of the data set
        period_mean, epoch_mean = np.polyfit(nr, tmids, 1)             
        
        fig, axes = plt.subplots(2,1,figsize=(6,8),tight_layout=True,sharex=True)
        axes[0].plot(nr, tmids, 'bo', label='Companion '+companion)
        axes[0].plot(nr, epoch_mean + nr * period_mean, 'b-')
        axes[0].set(xlabel='Nr.', ylabel='Transit mid-time')
        axes[0].legend()
        axes[1].plot(nr, tmids-(epoch_mean + nr * period_mean), 'bo')
        axes[1].axhline(0,c='grey',ls='--')
        fig.savefig(os.path.join(datadir,'ttv_preparation','ttv_preparation_'+companion+'_oc.pdf'), bbox_inches='tight')
        
        period_dev = np.abs( (period_mean-alles.initial_guess_params_median[companion+'_period'])/alles.initial_guess_params_median[companion+'_period'] )
        epoch_dev = np.abs( (epoch_mean-alles.initial_guess_params_median[companion+'_epoch'])/alles.initial_guess_params_median[companion+'_epoch'] )
        
        print('\nCompanion', companion)
        print('Initial guess for mean period and epoch:')
        print(np.format_float_positional(alles.initial_guess_params_median[companion+'_period']), 
              np.format_float_positional(alles.initial_guess_params_median[companion+'_epoch']))
        print('New estimate for mean period and epoch:')
        print(np.format_float_positional(period_mean,4), 
              np.format_float_positional(epoch_mean,4))
        # print('Deviation from another:')
        # print(np.format_float_positional(period_dev,4), 
        #       np.format_float_positional(epoch_dev,4))
        if (period_dev > 0.01) or (epoch_dev > 0.01):
            print('\n! Consider updating your initial guess to these new estimated mean values.')
            print('\n! If you do, then you must rerun this code.')
        else:
            print('\n! Looks great! You are ready to fit.')

        
        #----------------------------------------------------------------------
        #::: full lightcurve plot
        #----------------------------------------------------------------------
        flux_min = np.nanmin(all_flux)
        flux_max = np.nanmax(all_flux)
        if ax0 is None:
            days = np.max(all_times) - np.min(all_times)
            figsizex = np.max( [5, 5*(days/10.)] )
            fig, ax = plt.subplots(figsize=(figsizex, 4)) #figsize * 5 for every 20 days
        for inst in alles.settings['inst_phot']:
            ax.plot(alles.fulldata[inst]['time'], alles.fulldata[inst]['flux'],ls='none',marker='.',color='silver')
            # ax.plot(alles.data[inst]['time'], alles.data[inst]['flux'],ls='none',marker='.',label=inst) #color code by instrument
        plot_all_transits_color_coded() #color code by companion
                
        ax.plot( alles.data[companion+'_tmid_observed_transits'], np.ones_like(alles.data[companion+'_tmid_observed_transits'])*0.997*flux_min, 'k^', zorder=12 )
        for i, tmid in enumerate(alles.data[companion+'_tmid_observed_transits']):
            ax.text( tmid, 0.992*flux_min, str(i+1), ha='center', zorder=12 )  
            ax.axvline( tmid, color='grey', zorder=11 )
        ax.set(ylim=[0.99*flux_min, 1.002*flux_max], xlabel='Time (BJD)', ylabel='Realtive Flux', title='Companion '+companion) 
        ax.legend(loc='best')
    
        fname = os.path.join(datadir,'ttv_preparation','ttv_preparation_'+companion+'.jpg')
        fig = plt.gcf()
        fig.savefig(fname, bbox_inches='tight' )  
        plt.close(fig)
                